import shutil
import requests
from requests.auth import HTTPBasicAuth
from requests.exceptions import ConnectionError, ConnectTimeout, ReadTimeout
import keyring
import getpass

from jolt import utils
from jolt import cache
from jolt import log
from jolt import config
from jolt import filesystem as fs

NAME = "artifactory"
TIMEOUT = (3.5, 27)


class Artifactory(cache.StorageProvider):
    def __init__(self, cache):
        super(Artifactory, self).__init__()
        self._cache = cache
        self._uri = config.get(NAME, "uri")
        assert self._uri, "artifactory URI not configured"
        if self._uri[-1] != "/":
            self._uri += "/"
        self._repository = config.get(NAME, "repository", "jolt")
        self._upload = config.getboolean(NAME, "upload", True)
        self._download = config.getboolean(NAME, "download", True)
        self._disabled = False

    def _get_auth(self):
        service = config.get(NAME, "keyring.service")
        if not service:
            service = utils.read_input(NAME + " keyring service name (artifactory): ")
            if not service:
                service = NAME
            config.set(NAME, "keyring.service", service)
            config.save()

        username = config.get(NAME, "keyring.username")
        if not username:
            username = utils.read_input(NAME + " username: ")
            assert username, "no username configured for " + NAME
            config.set(NAME, "keyring.username", username)
            config.save()

        password = config.get(NAME, "keyring.password") or \
                   keyring.get_password(NAME, username)
        if not password:
            password = getpass.getpass(NAME + " password: ")
            assert password, "no password in keyring for " + NAME
            keyring.set_password(service, username, password)
        return HTTPBasicAuth(username, password)

    def _get_url(self, node, artifact):
        return "{uri}{repository}/{name}/{file}".format(
            uri=self._uri,
            repository=self._repository,
            name=node.name,
            file=fs.path.basename(artifact.get_archive_path()))

    @utils.retried.on_exception((ConnectionError, ReadTimeout))
    def download(self, node, force=False):
        if self._disabled:
            return False
        if not self._download and not force:
            return False
        with self._cache.get_artifact(node) as artifact:
            url = self._get_url(node, artifact)
            if node.tools.download(url, artifact.get_archive_path(), timeout=TIMEOUT):
                return True
        return False

    def download_enabled(self):
        return not self._disabled and self._download

    @utils.retried.on_exception((ConnectionError, ReadTimeout))
    def upload(self, node, force=False):
        if self._disabled:
            return True
        if not self._upload and not force:
            return True
        with self._cache.get_artifact(node) as artifact:
            url = self._get_url(node, artifact)
            archive = artifact.get_archive()
            return node.tools.upload(archive, url, auth=self._get_auth(), timeout=TIMEOUT)
        return False

    def upload_enabled(self):
        return not self._disabled and self._upload

    @utils.retried.on_exception((ConnectionError, ReadTimeout))
    def location(self, node):
        if self._disabled:
            return False
        with self._cache.get_artifact(node) as artifact:
            url = self._get_url(node, artifact)
            try:
                response = requests.head(url, stream=True, timeout=TIMEOUT)
            except ConnectTimeout as e:
                self._disabled = True
                log.warn("[ARTIFACTORY] failed to establish server connection, disabled")
                return False

            log.hysterical("[ARTIFACTORY] Head: {0}", url)
            log.hysterical("[ARTIFACTORY] Response: {0}", response.status_code)
            return url if response.status_code == 200 else ''
        return False


@cache.RegisterStorage
class ArtifactoryFactory(cache.StorageProviderFactory):
    @staticmethod
    def create(cache):
        log.verbose("Artifactory loaded")
        return Artifactory(cache)



import inspect
import imp
import copy
import glob
import os

from jolt.tasks import Task, Test, Resource
from jolt import filesystem as fs
from jolt import log
from jolt import utils
from jolt.plugins.ninja import CXXExecutable, CXXLibrary


@utils.Singleton
class JoltLoader(object):
    filename = "*.jolt"

    def __init__(self):
        self._tasks = []
        self._tests = []
        self._source = []
        self._path = None

    def _load_file(self, path):
        classes = []

        directory = fs.path.dirname(path)
        name, ext = fs.path.splitext(fs.path.basename(path))

        with open(path) as f:
            self._source.append(f.read())

        module = imp.load_source("joltfile_{0}".format(name), path)
        for name in module.__dict__:
            obj = module.__dict__[name]
            if inspect.isclass(obj):
                classes.append(obj)

        cls_exclude_list = [Task, Resource, CXXLibrary, CXXExecutable]

        tasks = [cls for cls in classes
                 if issubclass(cls, Task) and cls not in cls_exclude_list \
                 and not cls.__name__.startswith("_")]
        for task in tasks:
            task.name = task.name or task.__name__.lower()
            task.joltdir = directory
        self._tasks += tasks

        tests = [cls for cls in classes if issubclass(cls, Test) and cls is not Test \
                 and not cls.__name__.startswith("_")]
        for test in tests:
            test.name = test.name or test.__name__.lower()
            test.joltdir = directory
        self._tests += tests

        log.verbose("Loaded: {0}", path)

        return tasks

    def load(self, recursive=False):
        files = []
        path = os.getcwd()
        root = fs.path.normpath("/")
        while not files:
            files = glob.glob(fs.path.join(path, JoltLoader.filename))
            for file in files:
                self._load_file(file)
            if files:
                self._path = path
                break
            path = fs.path.dirname(path)
            if path == root:
                break
        return self._tasks, self._tests

    def get_sources(self):
        return "\n".join(self._source)

    @property
    def joltdir(self):
        return self._path



import datetime
import hashlib
import os

from jolt import utils
from jolt import log
from jolt import filesystem as fs
from jolt.tools import Tools

_providers = []

@utils.Singleton
class HashInfluenceRegistry(object):
    @staticmethod
    def Register(cls):
        _providers.append(cls)

    def __init__(self):
        self._providers = [provider() for provider in _providers]

    def register(self, provider):
        self._providers.append(provider)

    def apply_all(self, task, sha):
        for influence in self.get_strings(task):
            sha.update(influence.encode())
            log.hysterical("{0}: {1}", task.name, influence)

    def get_strings(self, task):
        content = []
        for provider in self._providers + task.influence:
            for line in str(provider.get_influence(task)).splitlines():
                content.append("Influence-{0}: {1}".format(provider.name, line))
        return content


class HashInfluenceProvider(object):
    name = "X"
    def get_influence(self, task):
        raise NotImplemented()


class TaskAttributeInfluence(HashInfluenceProvider):
    def __init__(self, attrib):
        self._attrib = attrib
        self.name = attrib.title()

    def get_influence(self, task):
        return getattr(task, self._attrib)


def attribute(name):
    """ Add task attribute value as hash influence.

    Args:
        name (str): Name of task class attribute/property.

    Example:

    .. code-block:: python

        from jolt import influence

        @influence.source("attribute")
        class Example(Task):
            attribute = False
    """

    def _decorate(cls):
        _old_init = cls.__init__
        def _init(self, *args, **kwargs):
            _old_init(self, *args, **kwargs)
            self.influence.append(TaskAttributeInfluence(name))
        cls.__init__ = _init
        return cls
    return _decorate


class TaskSourceInfluence(HashInfluenceProvider):
    def __init__(self, funcname, obj=None):
        self.name = "Source:" + funcname
        self.funcname = funcname
        self.obj = obj

    def get_influence(self, task):
        obj = self.obj or task
        return utils.sha1(task._get_source(getattr(obj, self.funcname)))


def source(name):
    """ Add function source code as hash influence.

    Args:
        name (str): Name of function in task class.

    Example:

    .. code-block:: python

        from jolt import influence

        @influence.source("method")
        class Example(Task):
            def method(self):
                return False

    """
    def _decorate(cls):
        _old_init = cls.__init__
        def _init(self, *args, **kwargs):
            _old_init(self, *args, **kwargs)
            self.influence.append(TaskSourceInfluence(name))
        cls.__init__ = _init
        return cls
    return _decorate


@HashInfluenceRegistry.Register
class TaskNameInfluence(HashInfluenceProvider):
    name = "Name"
    def get_influence(self, task):
        return task.name


@HashInfluenceRegistry.Register
class TaskParameterInfluence(HashInfluenceProvider):
    name = "Parameters"
    def get_influence(self, task):
        return ",".join(
            sorted(["{0}={1}".format(key, value)
                    for key, value in task._get_parameters().items()]))



class TaskDateInfluence(HashInfluenceProvider):
    name = "Date"

    def __init__(self, fmt):
        self.fmt = fmt

    def get_influence(self, task):
        now = datetime.datetime.now()
        return now.strftime(self.fmt)


def _date_influence(fmt):
    def _decorate(cls):
        _old_init = cls.__init__
        def _init(self, *args, **kwargs):
            _old_init(self, *args, **kwargs)
            self.influence.append(TaskDateInfluence(fmt))
        cls.__init__ = _init
        return cls
    return _decorate


yearly = _date_influence("%Y")
""" Add yearly hash influence.

If nothing else changes, the task is re-executed once every year.

Example:

    .. code-block:: python

        from jolt import influence

        @influence.yearly
        class Example(Task):

"""

weekly = _date_influence("%Y-%w")
""" Add weekly hash influence.

If nothing else changes, the task is re-executed once every week.

Example:

    .. code-block:: python

        from jolt import influence

        @influence.weekly
        class Example(Task):

"""

monthly = _date_influence("%Y-%m")
""" Add monthly hash influence.

If nothing else changes, the task is re-executed once every month.

Example:

    .. code-block:: python

        from jolt import influence

        @influence.monthly
        class Example(Task):

"""

daily = _date_influence("%Y-%m-%d")
""" Add daily hash influence.

If nothing else changes, the task is re-executed once every day.

Example:

    .. code-block:: python

        from jolt import influence

        @influence.daily
        class Example(Task):

"""

hourly = _date_influence("%Y-%m-%d %H")
""" Add hourly hash influence.

If nothing else changes, the task is re-executed once every hour.

Example:

    .. code-block:: python

        from jolt import influence

        @influence.hourly
        class Example(Task):

"""



class TaskEnvironmentInfluence(HashInfluenceProvider):
    name = "Environment"

    def __init__(self, variable):
        self.variable = variable

    def get_influence(self, task):
        return self.variable + "=" + os.environ.get(self.variable, "<unset>")


def environ(variable):
    """ Add environment variable hash influence.

    Args:
        variable (str): Name of an environment variable that will
            influence the hash of the task.

    Example:

    .. code-block:: python

        from jolt import influence

        @influence.environ("CFLAGS")
        class Example(Task):

    """
    def _decorate(cls):
        _old_init = cls.__init__
        def _init(self, *args, **kwargs):
            _old_init(self, *args, **kwargs)
            self.influence.append(TaskEnvironmentInfluence(variable))
        cls.__init__ = _init
        return cls

    return _decorate



class FileInfluence(HashInfluenceProvider):
    def __init__(self, path):
        self.path = path
        self.name = "File:" + fs.path.basename(path)

    def get_influence(self, task):
        sha = hashlib.sha1()
        with open(task.tools.expand_path(self.path), "rb") as f:
            for data in iter(lambda: f.read(4096), ''):
                sha.update(data)
        return sha.hexdigest()


def files(pathname):
    """ Add file content hash influence.

    Args:
        pathname (str): A pathname pattern used to find files that will
                influence the hash of the task
                The pattern may contain simple shell-style
                wildcards such as '*' and '?'. Note: files starting with a
                dot are not matched by these wildcards.

    Example:

    .. code-block:: python

        from jolt import influence

        @influence.files("*.cpp")
        class Example(Task):

    """
    def _decorate(cls):
        _old_init = cls.__init__
        def _init(self, *args, **kwargs):
            _old_init(self, *args, **kwargs)
            f = []
            with Tools(self, self.joltdir) as tools:
                f = tools.glob(pathname)
            f.sort()
            for i in f:
                self.influence.append(FileInfluence(i))
        cls.__init__ = _init
        return cls

    return _decorate



from xml.etree import ElementTree as ET
from xml.etree.ElementTree import Element
from xml.etree.ElementTree import ElementTree
from xml.dom import minidom

from jolt.utils import cached



class SubElement(object):
    def __init__(self, tag='', elem=None):
        super(SubElement, self).__init__()
        self._elem = elem if elem is not None else Element(tag)

    @property
    def attrib(self):
        return self._elem.attrib

    def get(self, *args, **kwargs):
        return self._elem.get(*args, **kwargs)

    def set(self, *args, **kwargs):
        return self._elem.set(*args, **kwargs)



class Attribute(object):
    def __init__(self, attribute, varname=None, child=False, values=None):
        self.attribute = attribute
        self.varname = varname if varname is not None else attribute.lower()
        self.child = child
        self.values = values

    def __call__(self, cls):
        def decorate(cls, attribute, varname, child, values):
            def _check_value(value, values):
                if values and value not in values:
                    raise ValueError('{0} is not one of {1}'.format(value, values))

            def attr_get(self):
                if attribute not in self.attrib:
                    return ''
                return self.get(attribute)

            def attr_set(self, value):
                if value is None:
                    try:
                        self.attrib.pop(attribute)
                    except:
                        pass
                    finally:
                        return
                _check_value(value, values)
                return self.set(attribute, value)

            def child_get(self):
                if not hasattr(self, '_'+varname):
                    return ''
                return getattr(self, '_'+varname).text

            def child_set(self, value):
                _check_value(value, values)
                if not value: return
                if not hasattr(self, '_'+varname):
                    e = SubElement(attribute)
                    self.append(e)
                    setattr(self, '_'+varname, e)
                getattr(self,'_'+varname).text = value

            if not child:
                setattr(cls, varname, property(attr_get, attr_set))
            else:
                setattr(cls, varname, property(child_get, child_set))
            return cls
        return decorate(cls, self.attribute, self.varname, self.child, self.values)


class Composition(object):
    def __init__(self, cls, name):
        self.cls = cls
        self.name = name if name is not None else attribute.lower()

    def __call__(self, cls):
        def decorate(cls, comp_cls, name):
            def create(self, *args, **kwargs):
                child = comp_cls(*args, **kwargs)
                self.append(child)
                return child

            @property
            @cached.instance
            def get(self):
                children = list(self.getroot()) if isinstance(self, ElementTree) else list(self)
                children = [n for n in children if n.tag == name]
                return [comp_cls(elem=child) for child in children]

            setattr(cls, 'create_' + name, create)
            setattr(cls, name + 's', get)
            return cls
        return decorate(cls, self.cls, self.name)



from copy import copy
from threading import RLock

from jolt.influence import *
from jolt.tools import Tools
from jolt import utils
from jolt import filesystem as fs
from jolt.plugins import git
from jolt.scheduler import *
from jolt.xmldom import *


@Attribute('name')
@Attribute('fetch')
class RepoRemote(SubElement):
    def __init__(self, elem=None):
        super(RepoRemote, self).__init__('remote', elem=elem)


@Attribute('revision')
@Attribute('remote')
@Attribute('sync-j')
class RepoDefault(SubElement):
    def __init__(self, elem=None):
        super(RepoDefault, self).__init__('default', elem=elem)


@Attribute('path')
@Attribute('name')
@Attribute('revision')
@Attribute('upstream')
@Attribute('remote')
class RepoProject(SubElement):
    def __init__(self, elem=None):
        super(RepoProject, self).__init__('project', elem=elem)

    def get_diff(self):
        with self.tools.cwd(self.path_or_name):
            return self.tools.run("git diff HEAD", output_on_error=True)
        assert False, "git command failed"

    def get_head(self):
        with self.tools.cwd(self.path_or_name):
            return self.tools.run("git rev-parse HEAD", output_on_error=True).strip()
        assert False, "git command failed"

    def get_remote_branches(self, commit):
        with self.tools.cwd(self.path_or_name):
            result = self.tools.run("git branch -r --contains {0}", commit, output_on_error=True)
            if not result:
                return []
            result = result.strip().splitlines()
            result = [line.strip() for line in result]
            return result
        assert False, "git command failed"

    def get_local_commits(self):
        with self.tools.cwd(self.path_or_name):
            result = self.tools.run("git rev-list HEAD ^@{{upstream}}", output_on_error=True)
            if not result:
                return []
            result = result.strip("\r\n ")
            return result.splitlines()
        assert False, "git command failed"

    def has_local_ref(self, ref):
        with self.tools.cwd(self.path_or_name):
            try:
                return True if self.tools.run("git show-ref {0}", ref, output=False) else False
            except:
                return False

    def get_remote_ref(self, commit, remote, pattern=None):
        with self.tools.cwd(self.path_or_name):
            result = self.tools.run(
                "git ls-remote {0}{1}",
                remote,
                " {0}".format(pattern) if pattern else "",
                output_on_error=True)
            if not result:
                return None
            result = result.strip().splitlines()
            result = [line.split("\t") for line in result]
            result = {line[0]: line[1] for line in result}
            return result.get(commit)
        assert False, "git command failed"

    @property
    def path_or_name(self):
        return self.path or self.name


@Composition(RepoRemote, "remote")
@Composition(RepoDefault, "default")
@Composition(RepoProject, "project")
class RepoManifest(ElementTree):
    def __init__(self, task, path):
        super(RepoManifest, self).__init__(element=Element('manifest'))
        self.path = path
        self.tools = task.tools

    def append(self, element):
        self.getroot().append(element)

    def get(self, key):
        self.getroot().get(key)

    def set(self, key, value):
        self.getroot().set(key, value)

    def parse(self, filename=".repo/manifest.xml"):
        with open(fs.path.join(self.tools.getcwd(), filename)) as f:
            root = ET.fromstring(f.read())
            self._setroot(root)
            for project in self.projects:
                project.tools = self.tools
            return self
        raise Exception("failed to parse xml file")

    def format(self):
        return minidom.parseString(ET.tostring(self.getroot())).toprettyxml(indent="  ")

    def write(self, filename):
        with open(filename, 'w') as f:
            f.write(self.format())

    def get_remote(self, project):
        if len(self.defaults) > 0:
            return project.remote or self.defaults[0].remote
        return project.remote

    def get_upstream(self, project):
        if len(self.defaults) > 0:
            return project.upstream or self.defaults[0].revision or "master"
        return project.upstream or "master"

    def assert_clean(self):
        for project in self.projects:
            assert not project.get_diff(), \
                "repo project '{0}' has local changes"\
                .format(project.path_or_name)

    def lock_revisions(self):
        for project in self.projects:
            head = project.get_head()
            if not project.get_remote_branches(head):
                remote_ref = project.get_remote_ref(
                    head, self.get_remote(project))
                assert remote_ref, \
                    "repo project '{0}' has unpublished commits"\
                    .format(project.path_or_name)
                head = remote_ref

            if not project.upstream:
                if project.revision and project.has_local_ref(project.revision):
                    project.upstream = project.revision
                else:
                    project.upstream = self.get_upstream(project)

            if not project.revision.startswith("refs/changes"):
                project.revision = head


_git_repos = {}
_git_repo_lock = RLock()

_repo_manifests = {}
_repo_manifest_lock = RLock()


class RepoInfluenceProvider(HashInfluenceProvider):
    name = "Repo"
    path = "."

    def __init__(self, path=None, include=None, exclude=None, network=True):
        self.path = path or self.__class__.path
        self.include = include
        self.exclude = exclude
        self.network = network

    def get_influence(self, task):
        self.tools = Tools(task, task.joltdir)
        try:
            manifest_path = fs.path.join(task.joltdir, task._get_expansion(self.path))
            manifest = RepoManifest(task, manifest_path)
            manifest.parse(fs.path.join(manifest_path, ".repo", "manifest.xml"))

            result = []
            for project in sorted(manifest.projects, key=lambda p: p.name):
                if self.include is not None and project.path_or_name not in self.include:
                    continue
                if self.exclude is not None and project.path_or_name in self.exclude:
                    continue

                with _git_repo_lock:
                    gip = _git_repos.get(project.path_or_name)
                    if gip is None:
                        gip = git.GitInfluenceProvider(project.path_or_name)
                        _git_repos[project.path_or_name] = gip
                    result.append(gip.get_influence(task))

            return "\n".join(result)

        except KeyError as e:
            log.exception()
        assert False, "failed to calculate hash influence for repo manifest at {0}".format(self.path)

    def get_manifest(self, task):
        manifest_path = fs.path.join(task.joltdir, task._get_expansion(self.path))
        with _repo_manifest_lock:
            manifest = _repo_manifests.get(manifest_path)
            if manifest is None:
                manifest = RepoManifest(task, manifest_path)
                manifest.parse()
                manifest.assert_clean()
                manifest.lock_revisions()
                _repo_manifests[manifest_path] = manifest
            return manifest


class RepoNetworkExecutorExtension(NetworkExecutorExtension):
    def get_parameters(self, task):
        rip = list(filter(
            lambda n: isinstance(n, RepoInfluenceProvider),
            task.task.influence))
        if rip:
            assert len(rip) == 1, "task influenced by multiple repo manifests"
            rip = rip[0]
            if rip.network:
                manifest = rip.get_manifest(task.task)
                return {"repo_manifest": manifest.format()}
        return {}


@NetworkExecutorExtensionFactory.Register
class RepoNetworkExecutorExtensionFactory(NetworkExecutorExtensionFactory):
    def create(self):
        return RepoNetworkExecutorExtension()


def global_influence(path, include=None, exclude=None, network=True, cls=RepoInfluenceProvider):
    HashInfluenceRegistry.get().register(cls(path, include, exclude, network))


def influence(path='.', include=None, exclude=None, network=True, cls=RepoInfluenceProvider):
    def _decorate(taskcls):
        if "influence" not in taskcls.__dict__:
            taskcls.influence = copy(taskcls.influence)
        provider = cls(path, include, exclude, network)
        taskcls.influence.append(provider)
        return taskcls
    return _decorate



from .projects.pybuild import CXXToolchain as PamCXXToolchain
from .projects.toolchain import ToolchainExtender
from .tools.gnu import GNUToolFactory

tools = GNUToolFactory()
gcc = PamCXXToolchain("linux-gcc")
tools.configure(gcc)




from jolt.cache import *


class EnvironmentVariable(ArtifactStringAttribute):
    def __init__(self, name):
        super(EnvironmentVariable, self).__init__(name)
        self._old_value = None

    def apply(self, task, artifact):
        self._old_value = task.tools.getenv(self.get_name())
        task.tools.setenv(self.get_name(), self.get_value())

    def unapply(self, task, artifact):
        if self._old_value:
            task.tools.setenv(self.get_name(), self._old_value)
        else:
            task.tools.setenv(self.get_name())
        self._old_value = None


class PathEnvironmentVariable(EnvironmentVariable):
    def __init__(self, name="PATH"):
        super(PathEnvironmentVariable, self).__init__(name)

    def append(self, value):
        if self.get_value():
            self.set_value(self.get_value() + fs.pathsep + value)
        else:
            self.set_value(value)

    def apply(self, task, artifact):
        self._old_value = task.tools.getenv(self.get_name())
        paths = self.get_value().split(fs.pathsep)
        paths = [fs.path.join(artifact.path, path) for path in paths]
        new_val = fs.pathsep.join(paths)
        if self._old_value:
            new_val = new_val + fs.pathsep + task.tools.getenv(self.get_name())
        task.tools.setenv(self.get_name(), new_val)


class EnvironmentVariableSet(ArtifactAttributeSet):
    def __init__(self):
        super(EnvironmentVariableSet, self).__init__()

    def create(self, name):
        if name == "PATH":
            return PathEnvironmentVariable(name)
        if name == "LD_LIBRARY_PATH":
            return PathEnvironmentVariable(name)
        if name == "PKG_CONFIG_PATH":
            return PathEnvironmentVariable(name)
        return EnvironmentVariable(name)


@ArtifactAttributeSetProvider.Register
class EnvironmentVariableSetProvider(ArtifactAttributeSetProvider):
    def create(self, artifact):
        setattr(artifact, "environ", EnvironmentVariableSet())

    def parse(self, artifact, content):
        if "environ" not in content:
            return

        for key, value in content["environ"].items():
            setattr(artifact.environ, key, value)

    def format(self, artifact, content):
        if "environ" not in content:
            content["environ"] = {}

        for key, value in artifact.environ.items():
            content["environ"][key] = str(value)

    def apply(self, task, artifact):
        artifact.environ.apply(task, artifact)

    def unapply(self, task, artifact):
        artifact.environ.unapply(task, artifact)



from jolt.tasks import *
from jolt.plugins import directory
from jolt.cache import *
from jolt.graph import *
from jolt.scheduler import *


log.verbose("SelfDeploy loaded")


_path = fs.path.dirname(__file__)
_path = fs.path.dirname(_path)
_path = fs.path.dirname(_path)


@directory.influence(_path, pattern="*.py")
class Jolt(Task):
    name = "jolt"

    def info(self, fmt, *args, **kwargs):
        log.verbose(fmt, *args, **kwargs)

    def publish(self, artifact, tools):
        with tools.cwd(_path):
            artifact.collect('README.rst')
            artifact.collect('setup.py')
            artifact.collect('jolt/*.py')
            artifact.collect('jolt/*.job')
            artifact.collect('jolt/*/*.py')
            artifact.collect('jolt/*/*/*.py')


class SelfDeployExtension(NetworkExecutorExtension):
    @utils.cached.instance
    def get_parameters(self, task):
        registry = TaskRegistry()
        registry.add_task_class(Jolt)
        acache = ArtifactCache.get()
        env = JoltEnvironment(cache=acache)
        gb = GraphBuilder(registry)
        dag = gb.build(["jolt"])
        task = dag.select(lambda graph, task: True)
        assert len(task) == 1, "too many selfdeploy tasks found"
        task = task[0]
        if not acache.is_available_remotely(task):
            duration = utils.duration()
            factory = LocalExecutorFactory()
            executor = LocalExecutor(factory, task, force_upload=True)
            executor.run(env)
        jolt_url = acache.location(task)
        assert jolt_url, "failed to selfdeploy jolt to remote cache"
        return { "jolt_url":  jolt_url, "jolt_identity": task.identity[:8] }


@NetworkExecutorExtensionFactory.Register
class SelfDeployExtensionFactory(NetworkExecutorExtensionFactory):
    def create(self):
        return SelfDeployExtension()



#!/usr/bin/python
import sys

from jolt import cli
from jolt import log


def main():
    try:
        cli.cli()
    except KeyboardInterrupt as e:
        log.warn("Interrupted by user")
        sys.exit(1)
    except Exception as e:
        log.exception(e)
        sys.exit(1)


if __name__ == "__main__":
    main()



from concurrent.futures import ThreadPoolExecutor, Future, as_completed
from functools import partial
import traceback
import sys
import inspect
try:
    import asyncio
    has_asyncio = True
except:
    has_asyncio = False


from jolt import cache
from jolt import log
from jolt import utils
from jolt.options import JoltOptions


class JoltEnvironment(object):
    def __init__(self, **kwargs):
        self.__dict__.update(kwargs)


class TaskQueue(object):
    def __init__(self, strategy):
        self.futures = {}
        self.strategy = strategy
        self._aborted = False

    def submit(self, cache, task):
        if self._aborted:
            return None

        env = JoltEnvironment(cache=cache)
        executor = self.strategy.create_executor(task)
        assert executor, "no executor can execute the task; "\
            "requesting a network build without proper configuration?"

        task.set_in_progress()
        future = executor.submit(env)
        self.futures[future] = task
        return future

    def wait(self):
        for future in as_completed(self.futures):
            task = self.futures[future]
            try:
                result = future.result()
            except Exception as error:
                log.exception()
                return task, error
            finally:
                del self.futures[future]
            return task, None
        return None, None

    def abort(self):
        self._aborted = True
        for future, task in self.futures.items():
            task.info("Execution cancelled")
            future.cancel()
        if len(self.futures):
            log.info("Waiting for tasks to finish, please be patient")
        self.strategy.executors.shutdown()

    def is_aborted(self):
        return self._aborted

    def in_progress(self, task):
        return task in self.futures.values()


class Executor(object):
    def __init__(self, factory):
        self.factory = factory

    def submit(self, env):
        return self.factory.submit(self, env)

    def is_aborted(self):
        return self.factory.is_aborted()

    def run(self, env):
        pass


class LocalExecutor(Executor):
    def __init__(self, factory, task, force_upload=False):
        super(LocalExecutor, self).__init__(factory)
        self.task = task
        self.force_upload = force_upload

    def run(self, env):
        if has_asyncio:
            loop = asyncio.SelectorEventLoop()
            asyncio.set_event_loop(loop)

        try:
            self.task.started()
            self.task.run(env.cache, self.force_upload)
        except Exception as e:
            self.task.failed()
            log.exception()
            raise e
        else:
            self.task.finished()
        return self.task


class NetworkExecutor(Executor):
    pass


class SkipTask(Executor):
    def __init__(self, factory, task, *args, **kwargs):
        super(SkipTask, self).__init__(factory, *args, **kwargs)
        self.task = task

    def run(self, env):
        self.task.skipped()
        for ext in self.task.extensions:
            ext.skipped()
        return self.task


class Downloader(Executor):
    def __init__(self, factory, task, *args, **kwargs):
        super(Downloader, self).__init__(factory, *args, **kwargs)
        self.task = task

    def _download(self, env, task):
        try:
            task.started("Download")
            assert env.cache.download(task), \
                "failed to download artifact of task '{0} ({1})'"\
                .format(task.qualified_name, task.identity[:8])
        except Exception as e:
            task.failed("Download")
            raise e
        else:
            task.finished("Download")

    def run(self, env):
        self._download(env, self.task)
        for ext in self.task.extensions:
            self._download(env, ext)
        return self.task


class Uploader(Executor):
    def __init__(self, factory, task, *args, **kwargs):
        super(Uploader, self).__init__(factory, *args, **kwargs)
        self.task = task

    def _upload(self, env, task):
        try:
            task.started("Upload")
            assert env.cache.upload(task), \
                "failed to upload artifact of task '{0} ({1})'"\
                .format(task.qualified_name, task.identity[:8])
        except Exception as e:
            task.failed("Upload")
            raise e
        else:
            task.finished("Upload")

    def run(self, env):
        self._upload(env, self.task)
        for ext in self.task.extensions:
            self._upload(env, ext)

        return self.task


@utils.Singleton
class ExecutorRegistry(object):
    executor_factories = []
    extension_factories = []

    def __init__(self, options=None):
        self._options = options or JoltOptions()
        self._factories = [factory(self._options) for factory in self.__class__.executor_factories]
        self._local_factory = LocalExecutorFactory(self._options)
        self._extensions = [factory().create() for factory in self.__class__.extension_factories]

    def shutdown(self):
        for factory in self._factories:
            factory.shutdown()

    def create_skipper(self, task):
        return SkipTask(self._local_factory, task)

    def create_downloader(self, task):
        return Downloader(self._local_factory, task)

    def create_uploader(self, task):
        return Uploader(self._local_factory, task)

    def create_local(self, task):
        return self._local_factory.create(task)

    def create_network(self, task):
        for factory in self._factories:
            executor = factory.create(task)
            if executor is not None:
                return executor
        return self._local_factory.create(task)

    def get_network_parameters(self, task):
        parameters = {}
        for extension in self._extensions:
            parameters.update(extension.get_parameters(task))
        return parameters


class NetworkExecutorExtensionFactory(object):
    @staticmethod
    def Register(cls):
        # assert cls is Factory
        ExecutorRegistry.extension_factories.insert(0, cls)
        return cls

    def create(self):
        raise NotImplemented()


class NetworkExecutorExtension(object):
    def get_parameters(self, task):
        return {}


class ExecutorFactory(object):
    @staticmethod
    def Register(cls):
        # assert cls is Factory
        ExecutorRegistry.executor_factories.insert(0, cls)
        return cls

    def __init__(self, max_workers=None):
        self.pool = ThreadPoolExecutor(max_workers=max_workers)
        self._aborted = False

    def is_aborted(self):
        return self._aborted

    def shutdown(self):
        self._aborted = True
        self.pool.shutdown()

    def create(self, task):
        raise NotImplemented()

    def _run(self, executor, env):
        try:
            if not self.is_aborted():
                executor.run(env)
        except KeyboardInterrupt:
            assert False, "Interrupted by user"

    def submit(self, executor, env):
        return self.pool.submit(partial(self._run, executor, env))



class LocalExecutorFactory(ExecutorFactory):
    def __init__(self, options=None):
        super(LocalExecutorFactory, self).__init__(max_workers=1)
        self._options = options or JoltOptions()

    def create(self, task):
        return LocalExecutor(self, task)


class NetworkExecutorFactory(ExecutorFactory):
    def __init__(self, *args, **kwargs):
        super(NetworkExecutorFactory, self).__init__(*args, **kwargs)


class ExecutionStrategy(object):
    def create_executor(self, task):
        raise NotImplemented()


class LocalStrategy(ExecutionStrategy):
    def __init__(self, executors, cache):
        self.executors = executors
        self.cache = cache

    def create_executor(self, task):
        if not task.is_cacheable():
            return self.executors.create_local(task)
        if task.is_available_locally(self.cache):
            return self.executors.create_skipper(task)
        if self.cache.download_enabled() and task.is_available_remotely(self.cache):
            return self.executors.create_downloader(task)
        return self.executors.create_local(task)


class DistributedStrategy(ExecutionStrategy):
    def __init__(self, executors, cache):
        self.executors = executors
        self.cache = cache

    def create_executor(self, task):
        if task.is_resource():
            return self.executors.create_local(task)

        if not task.is_cacheable():
            return self.executors.create_network(task)

        # Check remote availability first so that the availability of
        # remote storage providers is made known when checking if
        # artifacts can be downloaded or not.
        remote = task.is_available_remotely(self.cache)

        if not self.cache.download_enabled():
            if not task.is_available_locally(self.cache):
                return self.executors.create_local(task)
            else:
                return self.executors.create_skipper(task)

        if remote:
            if not task.is_available_locally(self.cache):
                return self.executors.create_downloader(task)
            else:
                return self.executors.create_skipper(task)

        if task.is_available_locally(self.cache):
            if self.cache.upload_enabled():
                if task.is_uploadable(self.cache):
                    return self.executors.create_uploader(task)
                else:
                    return self.executors.create_network(task)
            else:
                return self.executors.create_skipper(task)

        if task.is_fast():
            if not self.cache.upload_enabled():
                return self.executors.create_network(task)
            else:
                return self.executors.create_local(task)

        return self.executors.create_network(task)


class WorkerStrategy(ExecutionStrategy):
    def __init__(self, executors, cache):
        self.executors = executors
        self.cache = cache

    def create_executor(self, task):
        if task.is_resource():
            return self.executors.create_local(task)

        assert self.cache.upload_enabled(),\
            "artifact upload must be enabled for workers, fix configuration"

        if not task.is_cacheable():
            return self.executors.create_local(task)

        if task.is_available_locally(self.cache):
            if not task.is_available_remotely(self.cache):
                return self.executors.create_uploader(task)
            else:
                return self.executors.create_skipper(task)

            return self.executors.create_downloader(task)

        if not self.cache.download_enabled():
            return self.executors.create_local(task)

        if task.is_available_remotely(self.cache):
            return self.executors.create_downloader(task)

        return self.executors.create_local(task)



##############################################################################
#
# (C) 2016 - Robert Andersson - All rights reserved.
#
# This file and its contents are the property of Robert Andersson
# and may not be distributed, copied, or disclosed, in whole or in part,
# for any reason without written consent of the copyright holder.
#
##############################################################################


class Tool(object):
    def __init__(self):
        pass

    def transform(project, out_project, sources):
        pass



from jolt.tasks import *
from jolt.influence import *
from jolt.tools import *
from jolt.scheduler import *
from jolt.loader import JoltLoader
from jolt import utils
from jolt import filesystem as fs

try:
    import pygit2
    has_pygit = True
except:
    has_pygit = False


class GitInfluenceProvider(HashInfluenceProvider):
    name = "Tree"
    path = "."

    def __init__(self, path=None):
        super(GitInfluenceProvider, self).__init__()
        self.path = path or self.__class__.path

    def _get_path(self, task):
        p = task.tools.expand_path(self.path)
        return p

    def _get_relative_path(self, task):
        p = self._get_path(task)
        if p.startswith(task.joltdir):
            p = p[len(task.joltdir):]
            p = p[1:] if p[0] == fs.sep else p
        return p

    @utils.cached.instance
    def _get_tree_hash(self, task, sha="HEAD"):
        with task.tools.cwd(self._get_path(task)):
            return task.tools.run("git rev-parse {0}:./".format(sha), output_on_error=True)
        return ""

    @utils.cached.instance
    def _get_diff(self, task):
        with task.tools.cwd(self._get_path(task)):
            return task.tools.run("git diff HEAD .", output_on_error=True)
        return ""

    @utils.cached.instance
    def _get_diff_hash(self, task):
        return utils.sha1(self._get_diff(task))

    def get_influence(self, task):
        return "{0}:{1}:{2}".format(
            self._get_relative_path(task),
            self._get_tree_hash(task),
            self._get_diff_hash(task)[:8])


def global_influence(path, cls=GitInfluenceProvider):
    HashInfluenceRegistry.get().register(cls(path))


def influence(path, cls=GitInfluenceProvider):
    def _decorate(taskcls):
        if "influence" not in taskcls.__dict__:
            taskcls.influence = copy.copy(taskcls.influence)
        provider = cls(path=path)
        taskcls.influence.append(provider)
        return taskcls
    return _decorate


class Git(Resource, GitInfluenceProvider):
    """ Clones a Git repo.

    Also influences the hash of consuming tasks, causing tasks to
    be re-executed if the cloned repo is modified.

    """

    name = "git"
    url = Parameter(help="URL to the git repo to clone. Required.")
    sha = Parameter(required=False, help="Specific commit sha to be checked out. Optional.")
    path = Parameter(required=False, help="Local path where repo should be cloned.")

    def __init__(self, *args, **kwargs):
        super(Git, self).__init__(*args, **kwargs)
        self.path = str(self.path) or self._get_name()
        self.path = fs.path.join(JoltLoader.get().joltdir, self.path)
        self.influence.append(self)

    @utils.cached.instance
    def _get_name(self):
        repo = fs.path.basename(self.url.get_value())
        name, _ = fs.path.splitext(repo)
        return name

    @utils.cached.instance
    def _get_git_folder(self):
        return fs.path.join(self.path, ".git")

    def _clone(self):
        assert not fs.path.exists(self.path), \
            "destination folder '{0}' already exists but is not a git repo"\
            .format(self.path)
        depth = "--depth 1" if self.sha.is_unset() else ""
        log.info("Cloning into {0}", self.path)
        self.tools.run("git clone {0} {1} {2}", depth, self.url, self.path, output_on_error=True)
        assert fs.path.exists(self._get_git_folder()),\
            "failed to clone git repo '{0}'".format(self._get_name())

    @utils.cached.instance
    def _is_synced(self):
        with self.tools.cwd(self.path):
            return True if self.tools.run("git branch -r --contains HEAD", output_on_error=True) else False
        return True

    def _is_cloned(self):
        return fs.path.exists(self._get_git_folder())

    def acquire(self, artifact, env, tools):
        if not self.sha.is_unset():
            assert self._is_synced(),\
                "explicit sha requested but git repo '{0}' has local commits"\
                .format(self._get_name())
            assert not self._get_diff(self), \
                "explicit sha requested but git repo '{0}' has local changes"\
                .format(self._get_name())
            # Should be safe to do this now
            with self.tools.cwd(self.path):
                self.tools.run("git checkout {0}", self.sha, output_on_error=True)

    def get_influence(self, task):
        if isinstance(task, Git):
            if not self._is_cloned():
                self._clone()
            if not self.sha.is_unset():
                return self._get_tree_hash(self, self.sha.get_value())
        return super(Git, self).get_influence(task)

TaskRegistry.get().add_task_class(Git)


class GitNetworkExecutorExtension(NetworkExecutorExtension):
    """ Sanity check that a local git repo can be built remotely """

    def get_parameters(self, task):
        for child in task.children:
            if isinstance(child.task, Git):
                task = child.task
                if task._is_cloned() and task.sha.is_unset():
                    assert task._is_synced(),\
                        "local commit found in git repo '{0}'; "\
                        "push before building remotely"\
                        .format(task._get_name())
                    assert not task._get_diff(task),\
                        "local changes found in git repo '{0}'; "\
                        "commit and push before building remotely"\
                        .format(task._get_name())
        return {}


@NetworkExecutorExtensionFactory.Register
class GitNetworkExecutorExtensionFactory(NetworkExecutorExtensionFactory):
    def create(self):
        return GitNetworkExecutorExtension()



import os
import errno
import shutil
import zipfile
import tempfile
import tarfile


path = os.path
sep = os.sep
pathsep = os.pathsep


def makedirs(path):
    try:
        os.makedirs(path)
    except OSError as e:
        if e.errno != errno.EEXIST:
            raise

mkdtemp = tempfile.mkdtemp

def rename(old, new):
    return os.rename(old, new)

def move(src, dst):
    return shutil.move(src, dst)

def rmtree(path, ignore_errors=False):
    shutil.rmtree(path, ignore_errors)

def unlink(path, ignore_errors=False):
    try:
        os.unlink(path)
    except:
        if not ignore_errors:
            raise

def symlink(src, dest, *args, **kwargs):
    os.symlink(src, dest, *args, **kwargs)

def copy(src, dest, symlinks=False):
    if not path.exists(dest):
        if dest[-1] == os.sep:
            makedirs(dest)
            dest = path.join(dest, path.basename(src))
        else:
            makedirs(path.dirname(dest))
    else:
        if dest[-1] == os.sep:
            dest = path.join(dest, path.basename(src))

    if path.isdir(src):
        shutil.copytree(src, dest, symlinks=symlinks)
    else:
        shutil.copy2(src, dest)


def scandir(scanpath):
    return [os.path.join(path, f)
            for path, dirs, files in os.walk(scanpath)
            for f in files]

make_archive = shutil.make_archive

def make_archive(file, path, remove=False):
    shutil.make_archive(file, "gztar", root_dir=path)
    if remove:
        rmtree(path)
    return get_archive(path)

def extract_archive(file, path, remove=False):
    name, ext = os.path.splitext(file)
    if not ext:
        ext = ".tar.gz"
        file += ext
    if ext == ".tar.gz":
        with tarfile.open(file, 'r:gz') as tar:
            makedirs(path)
            tar.extractall(path)
        if remove:
            os.unlink(file)
        return
    assert False, "unsupported file extension: {0}".format(ext)


def get_archive(path):
    return path + ".tar.gz"



import keyring
import sys
import getpass
import inspect
import hashlib
from jinja2 import Template
import time
from requests.exceptions import ConnectionError, ReadTimeout

from jolt import config
from jolt import utils
from jolt import scheduler
from jolt.tps.jenkins import Jenkins
from jolt import log
from jolt import loader
from jolt import filesystem as fs

NAME = "jenkins"
TYPE = "Remote execution"
TIMEOUT = (3.5, 27)


@utils.Singleton
class JenkinsServer(object):
    def __init__(self):
        username, password = self._get_auth()
        self._server = Jenkins(config.get(NAME, "uri"), username, password, timeout=TIMEOUT)
        self.get_job_info = self._server.get_job_info
        self.get_build_info = self._server.get_build_info
        self.get_build_console_output = self._server.get_build_console_output
        self.get_queue_item = self._server.get_queue_item
        self.build_job = self._server.build_job
        self.stop_build = self._server.stop_build
        self.cancel_queue = self._server.cancel_queue
        try:
            self._ok = self._check_job()
        except:
            log.exception()
            log.warn("[JENKINS] failed to establish server connection, disabled")
            self._ok = False

    def _get_auth(self):
        service = config.get(NAME, "keyring.service")
        if not service:
            service = utils.read_input(NAME + " keyring service name (jenkins): ")
            if not service:
                service = NAME
            config.set(NAME, "keyring.service", service)
            config.save()

        username = config.get(NAME, "keyring.username")
        if not username:
            username = utils.read_input(NAME + " username: ")
            assert username, "no username configured for " + NAME
            config.set(NAME, "keyring.username", username)
            config.save()

        password = config.get(NAME, "keyring.password") or \
                   keyring.get_password(NAME, username)
        if not password:
            password = getpass.getpass(NAME + " password: ")
            assert password, "no password in keyring for " + NAME
            keyring.set_password(service, username, password)
        return username, password

    def _get_sha(self, data):
        sha = hashlib.sha1()
        sha.update(data.encode())
        return sha.hexdigest()

    def _get_job_template_path(self):
        default = fs.path.join(fs.path.dirname(__file__), "jenkins.job")
        return config.get(NAME, "template", default)

    def _load_job_template(self):
        with open(self._get_job_template_path()) as f:
            xml = f.read()
        return xml, self._get_sha(xml)

    def _check_job(self):
        template_xml, template_sha = self._load_job_template()
        self.job_name = "{0}-{1}".format(
            config.get(NAME, "job", "Jolt"),
            template_sha[:6])

        try:
            job_xml = self._get_job_config(self.job_name)
        except:
            self._create_job(self.job_name, template_xml, self._server.create_job)
            view = config.get(NAME, "view")
            if view:
                self._add_job_to_view(view, self.job_name)
            return True
        if self._get_sha(job_xml) != template_sha:
            assert self._create_job(
                self.job_name, template_xml, self._server.reconfig_job), \
                "[JENKINS] failed to change misconfigured job"
        return True

    def ok(self):
        return self._ok

    @utils.retried.on_exception((ConnectionError, ReadTimeout))
    def _create_job(self, name, job_template, func):
        template = Template(job_template)
        network_config = config.get("network", "config", "", expand=False)
        xml = template.render(config=network_config)
        func(name, xml)
        return True

    @utils.retried.on_exception((ConnectionError, ReadTimeout))
    def _add_job_to_view(self, view, job):
        self._server.add_job_to_view(view, job)

    @utils.retried.on_exception((ConnectionError, ReadTimeout))
    def _get_job_config(self, job):
        return self._server.get_job_config(job)


class JenkinsExecutor(scheduler.NetworkExecutor):
    def __init__(self, factory, task):
        super(JenkinsExecutor, self).__init__(factory)
        self.server = JenkinsServer.get()
        self.factory = factory
        self.task = task
        self.job = self.server.job_name

    @utils.retried.on_exception((ConnectionError, ReadTimeout))
    def _get_queue_item(self, queue_id):
        return self.server.get_queue_item(queue_id)

    @utils.retried.on_exception((ConnectionError, ReadTimeout))
    def _get_build_info(self, build_id):
        return self.server.get_build_info(self.job, build_id)

    @utils.retried.on_exception((ConnectionError, ReadTimeout))
    def _build_job(self, parameters):
        return self.server.build_job(self.job, parameters)

    @utils.retried.on_exception((ConnectionError, ReadTimeout))
    def _cancel_queue(self, queue_id):
        return self.server.cancel_queue(queue_id)

    @utils.retried.on_exception((ConnectionError, ReadTimeout))
    def _stop_build(self, build_id):
        return self.server.stop_build(self.job, build_id)

    @utils.retried.on_exception((ConnectionError, ReadTimeout))
    def _get_console_log(self, build_id):
        if not config.getboolean(NAME, "console", True):
            return
        logtext = self.server.get_build_console_output(self.job, build_id)
        for line in logtext.splitlines():
            if line.startswith("[ERROR]"):
                log.error(line[8:], log_context=self.task.identity[:8])
            elif line.startswith("[VERBOSE]"):
                log.verbose(line[10:], log_context=self.task.identity[:8])
            elif line.startswith("[HYSTERICAL]"):
                log.hysterical(line[13:], log_context=self.task.identity[:8])
            else:
                log.stdout(line, log_context=self.task.identity[:8])

    def _run(self, env):
        task = [self.task.qualified_name]
        task += [t.qualified_name for t in self.task.extensions]

        parameters = {
            "joltfile": loader.JoltLoader.get().get_sources(),
            "task": " ".join(task),
            "task_identity": self.task.identity[:8],
            "task_default": " ".join(["-d {0}".format(d) for d in self.factory.options.default])
        }
        parameters.update(scheduler.ExecutorRegistry.get().get_network_parameters(self.task))

        queue_id = self._build_job(parameters)

        log.verbose("[JENKINS] Queued {0}", self.task.qualified_name)

        queue_info = self._get_queue_item(queue_id)
        while not queue_info.get("executable"):
            assert not queue_info.get("cancelled"),\
            "[JENKINS] {0} failed with status CANCELLED".format(
                self.task.qualified_name)
            if self.is_aborted():
                self._cancel_queue(queue_id)
                assert False, "[JENKINS] execution cancelled"
            time.sleep(5)
            queue_info = self._get_queue_item(queue_id)

        log.verbose("[JENKINS] Executing {0}", self.task.qualified_name)

        build_id = queue_info["executable"]["number"]
        build_info = self._get_build_info(build_id)
        while build_info["result"] not in ["SUCCESS", "FAILURE", "ABORTED"]:
            if self.is_aborted():
                self._stop_build(build_id)
                assert False, "[JENKINS] execution cancelled"
            time.sleep(5)
            build_info = self._get_build_info(build_id)

        log.verbose("[JENKINS] Finished {0}", self.task.qualified_name)

        if build_info["result"] != "SUCCESS":
            self._get_console_log(build_id)

        assert build_info["result"] == "SUCCESS", \
            "[JENKINS] {1} failed with status {0}".format(
                build_info["result"], self.task.qualified_name)

        assert env.cache.is_available_remotely(self.task), \
            "[JENKINS] no artifact produced for {0}, check configuration"\
            .format(self.task.qualified_name)

        assert env.cache.download(self.task) or \
            not env.cache.download_enabled(), \
            "[JENKINS] failed to download artifact for {0}"\
            .format(self.task.qualified_name)

        for extension in self.task.extensions:
            assert env.cache.download(extension) or \
                not env.cache.download_enabled(), \
                "[JENKINS] failed to download artifact for {0}"\
                .format(extension.qualified_name)

        return self.task

    def run(self, env):
        try:
            self.task.started(TYPE)
            for extension in self.task.extensions:
                extension.started(TYPE)
            self._run(env)
            for extension in self.task.extensions:
                extension.finished(TYPE)
            self.task.finished(TYPE)
        except Exception as e:
            log.exception()
            for extension in self.task.extensions:
                extension.failed(TYPE)
            self.task.failed(TYPE)
            raise e
        return self.task


@scheduler.ExecutorFactory.Register
class JenkinsExecutorFactory(scheduler.NetworkExecutorFactory):
    def __init__(self, options):
        workers = config.get(NAME, "workers", 16)
        super(JenkinsExecutorFactory, self).__init__(max_workers=workers)
        self._options = options

    @property
    def options(self):
        return self._options

    def create(self, task):
        server = JenkinsServer.get()
        if not server.ok():
            return None
        return JenkinsExecutor(self, task)

log.verbose("Jenkins loaded")



name = "jolt"

from .tasks import Task
from .tasks import Test
from .tasks import Resource
from .tasks import Parameter
from .tasks import BooleanParameter

from .cache import Artifact
from .cache import Context

from .tools import Tools



##############################################################################
#
# (C) 2016 - Robert Andersson - All rights reserved.
#
# This file and its contents are the property of Robert Andersson
# and may not be distributed, copied, or disclosed, in whole or in part,
# for any reason without written consent of the copyright holder.
#
##############################################################################


from build.transform import pybuild
from build.tools import Tool
import platform
from os import path


class Directory(pybuild.Command):
    def __init__(self, dirname):
        ignore_error = False
        if platform.system() == "Windows":
            cmdline = "cmd /c if not exist \"{dir}\" mkdir \"{dir}\"".format(dir=dirname)
            ignore_error = True
        else:
            cmdline = "mkdir -p \"{dir}\"".format(dir=dirname)
        info = ' [MKDIR] {}'.format(dirname)
        super(Directory, self).__init__(dirname, cmdline, info, ignore_error=ignore_error)

    @property
    def required(self):
        return not path.exists(self.product)

    @property
    def timestamp(self):
        return 0


class PyBuildDirectoryCreator(Tool):
    def __init__(self):
        super(PyBuildDirectoryCreator, self).__init__()

    def transform(self, cxx_project, dirname):
        job = cxx_project.get_job(dirname)
        if not job:
            job = Directory(dirname)
            cxx_project.add_job(job)
        return job



from copy import copy
import hashlib
import networkx as nx
from networkx.drawing.nx_agraph import write_dot
import traceback
from threading import RLock

from jolt.tasks import *
from jolt.utils import *
from jolt.influence import *
from jolt.tools import Tools
from jolt import log
from jolt import utils
from jolt import colors


class TaskProxy(object):
    def __init__(self, task, graph):
        self.task = task
        self.graph = graph
        self.children = []
        self.ancestors = []
        self.extensions = []

        self._extended_task = None
        self._in_progress = False
        self._completed = False

    def __hash__(self):
        return id(self)

    @property
    def tools(self):
        return self.task.tools

    @property
    def name(self):
        return self.task.name

    @property
    def canonical_name(self):
        return self.task.canonical_name

    @property
    def qualified_name(self):
        return utils.format_task_name(
            self.task.name,
            self.task._get_parameters())

    @property
    def short_qualified_name(self):
        return utils.format_task_name(
            self.task.name,
            self.task._get_explicitly_set_parameters())

    @property
    def log_name(self):
        return "({0} {1})".format(self.short_qualified_name, self.identity[:8])

    @property
    @cached.instance
    def identity(self):
        sha = hashlib.sha1()

        HashInfluenceRegistry.get().apply_all(self.task, sha)

        # print("{}: {}".format(self.name, [n.name for n in self.children]))
        for node in self.children:
            sha.update(node.identity.encode())

        if self._extended_task:
            sha.update(self._extended_task.identity.encode())

        self.task.identity = sha.hexdigest()
        return self.task.identity

    def __str__(self):
        return "{0}{1}".format(self.short_qualified_name, "*" if self.is_extension() else '')

    def info(self, fmt, *args, **kwargs):
        self.task.info(fmt + " " + self.log_name, *args, **kwargs)

    def warn(self, fmt, *args, **kwargs):
        self.task.warn(fmt + " " + self.log_name, *args, **kwargs)

    def error(self, fmt, *args, **kwargs):
        self.task.error(fmt + " " + self.log_name, *args, **kwargs)

    def has_children(self):
        return len(self.children) > 0

    def has_ancestors(self):
        return len(self.ancestors) > 0

    def is_cacheable(self):
        return self.task.is_cacheable()

    def is_resource(self):
        return isinstance(self.task, Resource)

    def has_artifact(self):
        return not self.is_resource()

    def has_extensions(self):
        return len(self.extensions) > 0

    def add_extension(self, task):
        if self.is_extension():
            self._extended_task.add_extension(task)
        else:
            self.extensions.append(task)

    def is_extension(self):
        return self._extended_task is not None

    def set_extended_task(self, task):
        self._extended_task = task

    def get_extended_task(self):
        if self.is_extension():
            return self._extended_task.get_extended_task()
        return self

    def is_available_locally(self, cache):
        tasks = [self] + self.extensions
        return all(map(cache.is_available_locally, tasks))

    def is_available_remotely(self, cache):
        tasks = [self] + self.extensions
        return all(map(cache.is_available_remotely, tasks))

    def is_uploadable(self, cache):
        tasks = [self] + self.extensions
        return all(map(cache.is_uploadable, tasks))

    def is_fast(self):
        tasks = [self.task] + [e.task for e in self.extensions]
        return all([task.fast for task in tasks])

    def in_progress(self):
        return self._in_progress

    def is_ready(self):
        if self.in_progress():
            return False

        if self.is_extension():
            return False

        return self.graph.is_leaf(self)

    def is_completed(self):
        return self._completed

    def set_in_progress(self):
        self._in_progress = True

    def finalize(self, dag):
        log.verbose("Finalizing: " + self.qualified_name)

        # Find all direct and transitive dependencies
        self.children = sorted(
            nx.descendants(dag, self),
            key=lambda t: t.qualified_name)

        # Exclude transitive resources dependencies
        self.children = list(
            filter(lambda n: not n.is_resource() or \
                   dag.are_neighbors(self, n),
                   self.children))

        self.anestors = nx.ancestors(dag, self)
        return self.identity

    def started(self, what="Execution"):
        self.task.info(colors.blue(what + " started " + self.log_name))
        self.duration = utils.duration()

    def failed(self, what="Execution"):
        self.error("{0} failed after {1}", what, self.duration)

    def finished(self, what="Execution"):
        assert not self._completed, "task has already been completed"
        self._completed = True
        try:
            self.graph.remove_node(self)
        except:
            self.warn("Pruned task was executed")
        self.task.info(colors.green(what + " finished after {0} " + self.log_name), self.duration)

    def skipped(self):
        self._completed = True
        try:
            self.graph.remove_node(self)
        except:
            self.warn("Pruned task was executed")

    def run(self, cache, force_upload=False, force_build=False):
        with self.tools:
            tasks = [self] + self.extensions
            available_locally = available_remotely = False, False

            for child in self.children:
                if not child.has_artifact():
                    continue
                if not cache.is_available_locally(child):
                    assert cache.download(child), \
                        "no artifact of task '{0} ({1})' available"\
                        .format(child.qualified_name, child.identity[:8])

            if not force_build:
                available_locally = all(map(cache.is_available_locally, tasks))
                if available_locally and not force_upload:
                    return
                available_remotely = all(map(cache.is_available_remotely, tasks))
                if not available_locally and available_remotely:
                    available_locally = cache.download(self)

            if force_build or not available_locally:
                with cache.get_context(self) as context:
                    with self.tools.cwd(self.task.joltdir):
                        self.task.run(context, self.tools)

                if cache.is_available_locally(self):
                    with cache.get_artifact(self) as artifact:
                        artifact.discard()

                with cache.get_artifact(self) as artifact:
                    with self.tools.cwd(self.task.joltdir):
                        self.task.publish(artifact, self.tools)
                    artifact.commit()

            if force_build or force_upload or not available_remotely:
                assert cache.upload(self, force=force_upload) or \
                    not cache.upload_enabled(), \
                    "Failed to upload artifact for {0}".format(self.name)

            for extension in self.extensions:
                try:
                    extension.started()
                    extension.run(cache, force_upload, force_build)
                except Exception as e:
                    extension.failed()
                    raise e
                else:
                    extension.finished()


class Graph(nx.DiGraph):
    def __init__(self):
        super(Graph, self).__init__()
        self._mutex = RLock()

    def remove_node(self, node):
        with self._mutex:
            super(Graph, self).remove_node(node)

    @property
    def tasks(self):
        with self._mutex:
            return [n for n in self.nodes]

    def has_tasks(self):
        with self._mutex:
            return len(self.nodes) > 0

    def prune(self, func):
        with self._mutex:
            for node in nx.topological_sort(self):
                if func(self, node):
                    log.hysterical("[GRAPH] Pruned {0} ({1})", node.name, node.identity)
                    self.remove_node(node)

            for node in self.nodes:
                log.hysterical("[GRAPH] Keeping {0} ({1})", node.qualified_name, node.identity)

    def select(self, func):
        with self._mutex:
            return [n for n in self.nodes if func(self, n)]

    def debug(self):
        with self._mutex:
            log.verbose("[GRAPH] Listing all nodes")
            for node in nx.topological_sort(self):
                log.verbose("[GRAPH]   " + node.qualified_name)

    def is_leaf(self, node):
        with self._mutex:
            return self.out_degree(node) == 0

    def is_root(self, node):
        with self._mutex:
            return self.in_degree(node) == 0

    def is_orphan(self, node):
        with self._mutex:
            return self.is_root(node) and self.is_leaf(node)

    def are_neighbors(self, n1, n2):
        with self._mutex:
            return n2 in self[n1]


class GraphBuilder(object):
    def __init__(self, registry):
        self.graph = Graph()
        self.nodes = {}
        self.registry = registry

    def _get_node(self, name):
        node = self.nodes.get(name)
        if not node:
            task = self.registry.get_task(name)
            node = self._build_node(TaskProxy(task, self.graph))
            self.nodes[name] = node
        return node

    def _build_node(self, node):
        self.graph.add_node(node)

        extended = node.task._get_extends()
        if extended:
            extended_node = self._get_node(extended)
            self.graph.add_edges_from([(node, extended_node)])
            node.set_extended_task(extended_node)
            extended_node.add_extension(node)
            parent = extended_node.get_extended_task()
        else:
            parent = node

        for requirement in node.task._get_requires():
            child = self._get_node(requirement)
            self.graph.add_edges_from([(parent, child)])

        node.finalize(self.graph)
        return node

    def build(self, task_list):
        proxies = [self._get_node(task) for task in task_list]
        assert nx.is_directed_acyclic_graph(self.graph), "cyclic graph"
        return self.graph

    def display(self):
        t = tools.Tools()
        with t.tmpdir("dot") as tmpdir, t.cwd(tmpdir.get_path()):
            write_dot(self.graph, fs.path.join(t.getcwd(), 'graph.dot'))
            t.run('dot -Tsvg graph.dot -o graph.svg')
            t.run('eog graph.svg')



import glob
import json
import copy
import os
from tempfile import mkdtemp
import time
from threading import RLock, current_thread

from jolt import filesystem as fs
from jolt import log
from jolt import tools
from jolt import influence
from jolt import config
from jolt import utils
from jolt.options import JoltOptions


DEFAULT_ARCHIVE_TYPE = ".tar.gz"

def locked(f):
    def _f(self, *args, **kwargs):
        with self._lock:
            return f(self, *args, **kwargs)
    return _f


class StorageProvider(object):
    def download(self, node, force=False):
        return False

    def download_enabled(self):
        return True

    def upload(self, node, force=False):
        return False

    def upload_enabled(self):
        return True

    def location(self, node):
        return ''  # URL


class StorageProviderFactory(StorageProvider):
    def create(self):
        pass

def RegisterStorage(cls):
    ArtifactCache.storage_provider_factories.append(cls)


class ArtifactAttributeSet(object):
    def __init__(self):
        super(ArtifactAttributeSet, self).__setattr__("_attributes", {})

    def _get_attributes(self):
        return self._attributes

    def __getattr__(self, name):
        attributes = self._get_attributes()
        if name not in attributes:
            attributes[name] = self.create(name)
        return attributes[name]

    def __setattr__(self, name, value):
        attributes = self._get_attributes()
        if name not in attributes:
            attributes[name] = self.create(name)
        attributes[name].set_value(value)
        return attributes[name]

    def __dict__(self):
        return {key: str(value) for key, value in self.items()}

    def items(self):
        return self._get_attributes().items()

    def apply(self, task, artifact):
        for _, value in self.items():
            value.apply(task, artifact)

    def unapply(self, task, artifact):
        for _, value in self.items():
            value.unapply(task, artifact)


class ArtifactAttributeSetRegistry(object):
    providers = []

    @staticmethod
    def create_all(artifact):
        for provider in ArtifactAttributeSetRegistry.providers:
            provider().create(artifact)

    @staticmethod
    def parse_all(artifact, content):
        for provider in ArtifactAttributeSetRegistry.providers:
            provider().parse(artifact, content)

    @staticmethod
    def format_all(artifact, content):
        for provider in ArtifactAttributeSetRegistry.providers:
            provider().format(artifact, content)

    @staticmethod
    def apply_all(task, artifact):
        for provider in ArtifactAttributeSetRegistry.providers:
            provider().apply(task, artifact)

    @staticmethod
    def unapply_all(task, artifact):
        for provider in ArtifactAttributeSetRegistry.providers:
            provider().unapply(task, artifact)


class ArtifactAttributeSetProvider(object):
    @staticmethod
    def Register(cls):
        ArtifactAttributeSetRegistry.providers.append(cls)

    def create(self, artifact):
        raise NotImplemented()

    def parse(self, artifact, content):
        raise NotImplemented()

    def format(self, artifact, content):
        raise NotImplemented()

    def apply(self, task, artifact):
        raise NotImplemented()

    def unapply(self, task, artifact):
        raise NotImplemented()


class ArtifactAttribute(object):
    def __init__(self, name):
        self._name = name

    def get_name(self):
        return self._name

    def set_value(self, value):
        raise NotImplemented()

    def get_value(self):
        raise NotImplemented()

    def apply(self, task, artifact):
        pass

    def unapply(self, task, artifact):
        pass

    def __str__(self):
        raise NotImplemented()


class ArtifactStringAttribute(ArtifactAttribute):
    def __init__(self, name):
        self._name = name
        self._value = None

    def get_name(self):
        return self._name

    def set_value(self, value):
        self._value = value

    def get_value(self):
        return self._value

    def apply(self, task, artifact):
        pass

    def unapply(self, task, artifact):
        pass

    def __str__(self):
        return str(self._value)


class Artifact(object):
    """
    An artifact is a collection of files and metadata produced by a task.

    Task implementors call artifact methods to collect files to be published.
    In addition to files, other metadata can be provided as well, such as variables
    that should be set in the environment of consumer tasks.

    """

    environ = {}
    """ Artifact environment variables.

    A task can add environment variables to an artifact. Such a
    variable will automatically be set in the environment when
    consumer tasks are executed. A common use-case is to add
    programs to the PATH.

    Values appended to PATH-type variables are relative to the artifact
    root. They will be automatically expanded to absolute paths. These
    PATH-type variables are supported:

    - ``PATH``
    - ``LD_LIBRARY_PATH``
    - ``PKG_CONFIG_PATH``

    Example:

        .. code-block:: python

            def publish(self, artifact, tools):
                artifact.environ.PATH.append("bin")
                artifact.environ.JAVA_HOME = artifact.final_path
    """

    strings = {}
    """ Artifact strings.

    A task can add arbitrary string values to an artifact. Such a
    string will be available for consumer tasks to read.

    Example:

        .. code-block:: python

            def publish(self, artifact, tools):
                artifact.strings.version = "1.2"
    """

    def __init__(self, cache, node):
        self._cache = cache
        self._node = node
        self._path = cache.get_path(node)
        self._stable_path = cache.get_stable_path(node)
        self._temp = cache.create_path(node) \
                     if not fs.path.exists(cache.get_path(node)) \
                     else None
        self._archive = None
        self._unpacked = False
        self._uploadable = True
        self._size = 0
        ArtifactAttributeSetRegistry.create_all(self)
        self._read_manifest()

    def __enter__(self):
        return self

    def __exit__(self, type, value, tb):
        if self._archive:
            fs.unlink(self._archive)
        if self._temp:
            fs.rmtree(self._temp)

    def __getattr__(self, name):
        assert False, "no attribute '{0}' in artifact for '{1} ({2})'"\
            .format(name, self._node.qualified_name, self._node.identity[:8])

    def _write_manifest(self):
        content = {}
        content["task"] = self._node.name
        content["size"] = self._get_size()
        content["unpacked"] = self._unpacked
        content["uploadable"] = self._uploadable
        content["identity"] = self._node.identity
        content["requires"] = self._node.task._get_requires()
        content["parameters"] = self._node.task._get_parameters()
        content["influence"] = influence.HashInfluenceRegistry.get().get_strings(self._node.task)
        ArtifactAttributeSetRegistry.format_all(self, content)

        manifest = fs.path.join(self._temp, ".manifest.json")
        with open(manifest, "wb") as f:
            f.write(json.dumps(content, indent=3).encode())

    def _read_manifest(self):
        if self._temp:
            return
        manifest = fs.path.join(self._path, ".manifest.json")
        with open(manifest, "rb") as f:
            content = json.loads(f.read().decode())
            self._size = content["size"]
            self._unpacked = content["unpacked"]
            self._uploadable = content["uploadable"]
            ArtifactAttributeSetRegistry.parse_all(self, content)

    def _get_size(self):
        counted_inodes = {}
        size = 0
        for path, dirs, files in os.walk(self.path):
            for file in files:
                fp = os.path.join(path, file)
                try:
                    stat = os.lstat(fp)
                except OSError:
                    continue

                try:
                    counted_inodes[stat.st_ino]
                except KeyError:
                    counted_inodes[stat.st_ino] = True
                else:
                    continue

                size += stat.st_size

            for dir in dirs:
                fp = os.path.join(path, dir)
                try:
                    stat = os.lstat(fp)
                except OSError:
                    continue

                size += stat.st_size
        return size

    def apply(self):
        fs.unlink(self._stable_path, ignore_errors=True)
        os.symlink(self._path, self._stable_path)

    def unapply(self):
        fs.unlink(self._stable_path, ignore_errors=True)

    def commit(self, uploadable=True):
        if not self._node.task.is_cacheable():
            return
        if self._temp:
            self._uploadable = uploadable
            self._size = self._get_size()
            self._write_manifest()
            fs.rename(self._temp, self._path)
            self._temp = None
            self._cache.commit(self)

    def discard(self):
        if self._archive:
            fs.unlink(self._archive)
        if self._temp:
            fs.rmtree(self._temp)
        if self._path:
            fs.rmtree(self._path)

    def modify(self):
        assert not self._temp, \
            "attempting to unpack absent artifact of task '{0} ({1})'"\
            .format(self._node.qualified_name, self._node.identity[:8])
        #self._temp = self._cache.get_path(self._node) + ".unpack"
        #fs.move(self._path, self._temp)
        self._temp = self._path
        self._unpacked = True

    @property
    def path(self):
        """ str: The current location of the artifact in the local cache. """
        return self._temp or self._path

    @property
    def final_path(self):
        """ str: The final location of the artifact in the local cache. """
        return self._path

    @property
    def stable_path(self):
        """ str: A stable location of the artifact in the local cache. """
        return self._stable_path

    @property
    def tools(self):
        return self._node.tools

    def collect(self, files, dest=None, flatten=False, symlinks=False):
        """ Collect files to be included in the artifact.

        Args:
            files (str): A filename pattern matching the files to be include
                in the artifact. The pattern may contain simple shell-style
                wildcards such as '*' and '?'. Note: files starting with a
                dot are not matched by these wildcards.
            dest (str, optional): Destination path within the artifact. If
                the string ends with a path separator a new directory will
                be created and all matched source files will be copied into
                the new directory. A destination without trailing path
                separator can be used to rename single files, one at a time.
            flatten (boolean, optional): If True, the directory tree structure
                of matched source files will flattened, i.e. all files will
                be copied into the root of the destination path. The default
                is False, which retains the directory tree structure
                relative to the current working directory.
            symlinks (boolean, optional): If True, symlinks are copied.
                The default is False, i.e. the symlink target is copied.

        """

        assert self._temp, \
            "can't collect files into already published artifact of task '{0} ({1})'"\
            .format(self._node.qualified_name, self._node.identity[:8])
        files = self._node.task._get_expansion(files)
        dest = self._node.task._get_expansion(dest) if dest is not None else None

        files = self.tools.glob(files)
        dirname = fs.path.join(self._temp, dest) if dest else self._temp + fs.sep
        for src in files:
            srcs = fs.scandir(src) if fs.path.isdir(src) and flatten else [src]
            for src in srcs:
                dest = fs.path.join(dirname, src) \
                       if not flatten else \
                          fs.path.join(dirname, fs.path.basename(src))
                self.tools.copy(src, dest, symlinks=symlinks)
                log.verbose("Collected {0} -> {1}", src, dest[len(self._temp):])

    def copy(self, pathname, dest, flatten=False, symlinks=False):
        """ Copy files from the artifact.

        Args:
            pathname (str): A pathname pattern, relative to the root, matching
                the files to be copied from the artifact.
                The pattern may contain simple shell-style
                wildcards such as '*' and '?'. Note: files starting with a
                dot are not matched by these wildcards.
            dest (str, optional): Destination path. If the string ends with a
                path separator a new directory will
                be created and all matched source files will be copied into
                the new directory. A destination without trailing path
                separator can be used to rename single files, one at a time.
            flatten (boolean, optional): If True, the directory tree structure
                of matched source files will flattened, i.e. all files will
                be copied into the root of the destination path. The default
                is False, which retains the directory tree structure.
            symlinks (boolean, optional): If True, symlinks are copied.
                The default is False, i.e. the symlink target is copied.

        """

        assert self._temp, \
            "can't copy files from the unpublished artifact of task '{0} ({1})'"\
            .format(self._node.qualified_name, self._node.identity[:8])
        pathname = self._node.task._get_expansion(pathname)
        dest = self._node.task._get_expansion(dest)

        files = []
        with self.tools.cwd(self._path):
            files = self.tools.glob(pathname)
        for src in files:
            srcs = fs.scandir(src) if fs.path.isdir(src) and flatten else [src]
            for src in srcs:
                destfile = fs.path.join(dest, src) \
                           if not flatten else \
                              fs.path.join(dest, fs.path.basename(src))
                self.tools.copy(fs.path.join(self._path, src), destfile, symlinks=symlinks)
                log.verbose("Copied {0} -> {1}", src, destfile)

    def compress(self):
        assert not self._temp, \
            "can't compress the unpublished artifact of task '{0} ({1})'"\
            .format(self._node.qualified_name, self._node.identity[:8])
        if not self.get_archive():
            self._archive = self.tools.archive(
                self._path, self._path + DEFAULT_ARCHIVE_TYPE)

    def decompress(self):
        archive = self._path + DEFAULT_ARCHIVE_TYPE
        self.tools.extract(archive, self._path)
        self.tools.unlink(archive)
        if self._temp:
            fs.rmtree(self._temp)
            self._temp = None
        self._read_manifest()

    def get_archive(self):
        path = fs.get_archive(self._path)
        return path if fs.path.exists(path) else None

    def get_archive_path(self):
        return fs.get_archive(self._path)

    def get_size(self):
        return self._size

    def get_task(self):
        return self._node.task

    def get_cache(self):
        return self._cache

    def get_identity(self):
        return self._node.identity

    def is_temporary(self):
        return self._temp is not None

    def is_unpacked(self):
        return self._unpacked

    def is_uploadable(self):
        return self._uploadable


class Context(object):
    """
    Execution context and dependency wrapper.

    A ``Context`` gathers dependencies and initializes the environment
    for an executing task. It is passed as an argument to the
    Task's :func:`~jolt.Task.run` method.

    A task implementor can use the context as a dictionary of dependencies where
    the key is the name of a dependency and the value is the dependency's
    :class:`~jolt.Artifact`.

    """

    def __init__(self, cache, node):
        self._cache = cache
        self._node = node
        self._artifacts = {}
        self._artifacts_index = {}

    def __enter__(self):
        for dep in self._node.children:
            self._cache.unpack(dep)
            with self._cache.get_artifact(dep) as artifact:
                self._artifacts[dep.qualified_name] = artifact
                self._artifacts_index[dep.qualified_name] = artifact
                self._artifacts_index[dep.short_qualified_name] = artifact
                artifact.apply()
                ArtifactAttributeSetRegistry.apply_all(self._node.task, artifact)
        return self

    def __exit__(self, type, value, tb):
        for name, artifact in self._artifacts.items():
            ArtifactAttributeSetRegistry.unapply_all(self._node.task, artifact)
            artifact.unapply()

    def __getitem__(self, key):
        """ Get artifact for a task listed as a requirement.

        Args:
            key (str): Name of the task listed as a requirement.

        Returns:
            The :class:`~jolt.Artifact` associated with the task.

        Example:
            .. code-block:: python

                requires = "dependency"

                def run(self, deps, tools):
                    dependency_artifact = deps["dependency"]

        """

        key = self._node.task._get_expansion(key)
        assert key in self._artifacts_index, \
            "no such dependency '{0}' for task '{1} ({2})'"\
            .format(key, self._node.qualified_name, self._node.identity[:8])
        return self._artifacts_index[key]

    def items(self):
        """ List all requirements and their artifacts.

        Returns:
            Requirement dictionary items. Each item is a tuple with the
            requirement name and the artifact.
        """
        return self._artifacts.items()


class CacheStats(object):
    def __init__(self, cache):
        self._lock = RLock()
        self.cache = cache
        self.path = fs.path.join(cache.root, ".stats.json")
        try:
            self.load()
        except:
            self.stats = {}
        self.active = set()
        log.verbose("Cache size is {0}", utils.as_human_size(self.get_size()))

    def load(self):
        with open(self.path) as f:
            self.stats = json.loads(f.read().decode())

        deleted = []
        for artifact, stats in self.stats.items():
            path = fs.path.join(self.cache.root, stats["name"], artifact)
            if not os.path.exists(path):
                deleted.append(artifact)
        for deleted in deleted:
            del self.stats[deleted]

        self.save()

    @locked
    def save(self):
        with open(self.path, "wb") as f:
            f.write(json.dumps(self.stats, indent=3).encode())

    @locked
    def update(self, artifact, save=True):
        if artifact.is_temporary():
            return
        stats = {}
        stats["name"] = artifact.get_task().canonical_name
        stats["used"] = time.time()
        stats["size"] = artifact.get_size()
        self.stats[artifact.get_identity()] = stats
        self.active.add(artifact.get_identity())
        if save:
            self.save()

    @locked
    def remove(self, artifact):
        del self.stats[artifact["identity"]]
        self.save()

    @locked
    def get_size(self):
        size = 0
        for artifact, stats in self.stats.items():
            size += stats["size"]
        return size

    @locked
    def get_lru(self):
        assert self.stats, "can't evict LRU artifact, no artifacts in cache"
        nt = [dict(identity=artifact, **stats) for artifact, stats in self.stats.items()]
        nt = sorted(nt, key=lambda x: x["used"])
        # Don't evict artifacts in the current working set
        nt = list(filter(lambda x: x["identity"] not in self.active, nt))
        return nt[0] if len(nt) > 0 else None


@utils.Singleton
class ArtifactCache(StorageProvider):
    root = fs.path.join(fs.path.expanduser("~"), ".cache", "jolt")
    storage_provider_factories = []

    def __init__(self, options=None):
        try:
            self.root = config.get("jolt", "cachedir") or ArtifactCache.root
            fs.makedirs(self.root)
        except:
            assert False, "failed to create cache directory"

        self.max_size = config.getsize(
            "jolt", "cachesize", os.environ.get("JOLT_CACHESIZE", 1*1024**3))
        self.stats = CacheStats(self)
        self.storage_providers = [
            factory.create(self)
            for factory in ArtifactCache.storage_provider_factories]
        self._options = options or JoltOptions()

    def get_path(self, node):
        return fs.path.join(self.root, node.canonical_name, node.identity)

    def get_stable_path(self, node):
        identity = utils.sha1(node.qualified_name)
        return fs.path.join(self.root, node.canonical_name, identity)

    def evict(self):
        while self.stats.get_size() > self.max_size:
            artifact = self.stats.get_lru()
            if artifact is None:
                return
            log.verbose("Evicting artifact '{name}:{identity}'".format(**artifact))
            path = fs.path.join(self.root, artifact["name"], artifact["identity"])
            self.stats.remove(artifact)
            fs.rmtree(path, ignore_errors=True)

    def create_path(self, node):
        path = None
        try:
            dirname = fs.path.join(self.root, node.canonical_name)
            fs.makedirs(dirname)
            path = mkdtemp(prefix=node.identity, dir=dirname)
        except:
            pass
        assert path, "couldn't create temporary artifact directory for task '{0} ({1})'"\
            .format(node.qualified_name, node.identity[:8])
        return path

    def is_available_locally(self, node):
        if not node.task.is_cacheable():
            return False
        if fs.path.exists(self.get_path(node)):
            with self.get_artifact(node) as a:
                self.stats.update(a)
            return True
        return False

    def is_available_remotely(self, node):
        if not self.upload_enabled() and not self.download_enabled():
            return True
        if not node.task.is_cacheable():
            return False
        for provider in self.storage_providers:
            if provider.location(node):
                return True
        return False

    def is_available(self, node):
        return self.is_available_locally(node) or self.is_available_remotely(node)

    def is_uploadable(self, node):
        with self.get_artifact(node) as artifact:
            return artifact.is_uploadable()

    def download_enabled(self):
        return self._options.download and \
            any([provider.download_enabled() for provider in self.storage_providers])

    def download(self, node, force=False):
        if not force and not self.download_enabled():
            return False
        if not node.task.is_cacheable():
            return False
        if self.is_available_locally(node):
            node.info("Download skipped, already in local cache")
            return True
        for provider in self.storage_providers:
            if provider.download(node, force):
                with self.get_artifact(node) as artifact:
                    artifact.decompress()
                    self.commit(artifact)
                return True
        return len(self.storage_providers) == 0

    def upload_enabled(self):
        return self._options.upload and \
            any([provider.upload_enabled() for provider in self.storage_providers])

    def upload(self, node, force=False):
        if not force and not self.upload_enabled():
            return False
        if not node.task.is_cacheable():
            return True
        assert self.is_available_locally(node), \
            "can't upload artifact of task '{0} ({1})', absent in the local cache"\
            .format(node.qualified_name, node.identity[:8])
        if self.storage_providers:
            with self.get_artifact(node) as artifact:
                artifact.compress()
                return all([provider.upload(node, force) for provider in self.storage_providers])
        return len(self.storage_providers) == 0

    def location(self, node):
        if not node.task.is_cacheable():
            return ''
        for provider in self.storage_providers:
            url = provider.location(node)
            if url:
                return url
        return ''

    def unpack(self, node):
        if not node.task.is_cacheable():
            return False
        with self.get_artifact(node) as artifact:
            if artifact.is_unpacked():
                return True
            artifact.modify()
            task = artifact.get_task()
            with tools.Tools(task) as t:
                try:
                    task.unpack(artifact, t)
                    artifact.commit(uploadable=False)
                except NotImplementedError:
                    artifact.commit()
        return True

    def commit(self, artifact):
        self.stats.update(artifact)
        self.evict()

    def discard(self, node):
        if not self.is_available_locally(node):
            return False
        with self.get_artifact(node) as artifact:
            self.stats.remove(dict(identity=node.identity))
            artifact.discard()
        return True

    def get_context(self, node):
        return Context(self, node)

    def get_artifact(self, node):
        artifact = Artifact(self, node)
        self.stats.update(artifact)
        return artifact

    def get_archive_path(self, node):
        return fs.get_archive(self.get_path(node))

    def advise(self, node_list):
        """ Advise the cache about what artifacts to retain in the cache. """
        for node in node_list:
            if not node.task.is_cacheable():
                continue
            if fs.path.exists(self.get_path(node)):
                with Artifact(self, node) as artifact:
                    self.stats.update(artifact, save=False)
        self.stats.save()



from __future__ import print_function
import sys
import tqdm
import traceback
from datetime import datetime

from jolt import config
from jolt import filesystem as fs
from jolt import colors

ERROR = 0
WARN = 1
NORMAL = 2
VERBOSE = 3
HYSTERICAL = 4

_levelstr = ["ERROR", "WARNING", "INFO", "VERBOSE", "HYSTERICAL"]

default_path = fs.path.join(fs.path.expanduser("~"), ".jolt", "jolt.log")
path = config.get("jolt", "logfile", default_path)

_loglevel = NORMAL
try:
    dirpath = fs.path.dirname(path)
    if not fs.path.exists(dirpath):
        fs.makedirs(dirpath)
    _file = open(path, "a")
except:
    print("[ERROR] " + colors.red("could not open logfile: {0}".format(path)))
    print("[ERROR] " + colors.red("please set 'jolt.logfile' to an alternate path"))
    sys.exit(1)


def _prefix(level, **kwargs):
    if type(level) == int:
        level = "[{}]".format(_levelstr[level])
    elif type(level) == str:
        level = "[{}]".format(level)
    else:
        level = ""
    context = kwargs.get("log_context")
    context = "[{}]".format(context) if context else ""
    pad = " " if level and context else ""
    return context + pad + level + " "

def _line(level, fmt, *args, **kwargs):
    from jolt.utils import expand
    return _prefix(level, **kwargs) + expand(fmt, *args, ignore_errors=True, **kwargs)

def _streamwrite(stream, line):
    stream.write(line + "\r\n")
    stream.flush()

def _streamwrite_file(stream, line):
    ts = datetime.now().strftime("%Y-%m-%d %H:%M:%S.%f")
    _streamwrite(stream, ts + " " + line)

def _log(level, stream, fmt, *args, **kwargs):
    line = _line(level, fmt, *args, **kwargs)
    if level <= _loglevel:
        _streamwrite(stream, line)
    _streamwrite_file(_file, line)

def set_level(level):
    global _loglevel
    _loglevel = level

def info(fmt, *args, **kwargs):
    _log(NORMAL, sys.stdout, fmt, *args, **kwargs)

def warn(fmt, *args, **kwargs):
    _log(WARN, sys.stdout, colors.yellow(fmt), *args, **kwargs)

def verbose(fmt, *args, **kwargs):
    _log(VERBOSE, sys.stdout, fmt, *args, **kwargs)

def hysterical(fmt, *args, **kwargs):
    _log(HYSTERICAL, sys.stdout, fmt, *args, **kwargs)

def error(fmt, *args, **kwargs):
    _log(ERROR, sys.stdout, colors.red(fmt), *args, **kwargs)

def exception(exc=None):
    if exc:
        _streamwrite(sys.stderr, "[ERROR] " + colors.red(str(exc)))
        _streamwrite_file(_file, "[ERROR] " + str(exc))
    backtrace = traceback.format_exc()
    for line in backtrace.splitlines():
        _streamwrite_file(_file, "[ERROR] " + line)

def stdout(fmt, *args, **kwargs):
    try:
        line = utils.expand(fmt, *args, ignore_errors=True, **kwargs)
    except:
        line = fmt
    _streamwrite(sys.stdout, _prefix(None, **kwargs) + line)
    _streamwrite_file(_file, _prefix("STDOUT", **kwargs) + line)

def stderr(fmt, *args, **kwargs):
    try:
        line = utils.expand(fmt, *args, ignore_errors=True, **kwargs)
    except:
        line = fmt
    _streamwrite(sys.stderr, line)
    _streamwrite_file(_file, "[STDERR] " + line)

def progress(desc, count, unit):
    _streamwrite_file(_file, "[INFO] " + desc)
    p = tqdm.tqdm(total=count, unit=unit, unit_scale=True)
    p.set_description(desc)
    return p


_file.write("================================================================================\n")
_file.flush()



from jolt.xmldom import *
from jolt import filesystem as fs


@Attribute('name')
@Attribute('identity')
class _JoltTask(SubElement):
    def __init__(self, elem=None):
        super(_JoltTask, self).__init__('task', elem=elem)


@Composition(_JoltTask, "task")
class JoltManifest(ElementTree):
    def __init__(self):
        super(JoltManifest, self).__init__(element=Element('jolt-manifest'))

    def append(self, element):
        self.getroot().append(element)

    def get(self, key):
        self.getroot().get(key)

    def set(self, key, value):
        self.getroot().set(key, value)

    def parse(self, filename="default.joltmanifestx"):
        with open(filename) as f:
            data = f.read().replace("\n  ", "")
            data = data.replace("\n", "")
            root = ET.fromstring(data)
            self._setroot(root)
            return self
        raise Exception("failed to parse xml file")

    def format(self):
        return minidom.parseString(ET.tostring(self.getroot())).toprettyxml(indent="  ")

    def write(self, filename):
        with open(filename, 'w') as f:
            f.write(self.format())

    def has_task(self, task):
        return len(self.getroot().findall(".//task[@identity='{}']".format(task.identity))) != 0



import subprocess
import os
import threading
import sys
import threading
import termios
import glob
import multiprocessing
import shutil
import requests
import tarfile
import zipfile
import bz2file
from contextlib import contextmanager

from jolt import filesystem as fs
from jolt import log
from jolt import utils


def _run(cmd, cwd, env, *args, **kwargs):
    output = kwargs.get("output")
    output_on_error = kwargs.get("output_on_error")
    output = output if output is not None else True
    output = False if output_on_error else output
    p = subprocess.Popen(
        cmd,
        stdin=subprocess.PIPE,
        stdout=subprocess.PIPE,
        stderr=subprocess.PIPE,
        shell=True,
        cwd=cwd,
        env=env)

    class Reader(threading.Thread):
        def __init__(self, stream, output=None):
            super(Reader, self).__init__()
            self.output = output
            self.stream = stream
            self.buffer = []
            self.start()

        def run(self):
            try:
                for line in iter(self.stream.readline, b''):
                    line = line.rstrip()
                    line = line.decode(errors='ignore')
                    if self.output:
                        self.output(line)
                    self.buffer.append(line)
            except Exception as e:
                if self.output:
                    self.output("{0}", str(e))
                    self.output(line)
                self.buffer.append(line)

    stdout = Reader(p.stdout, output=log.stdout if output else None)
    stderr = Reader(p.stderr, output=log.stderr if output else None)
    p.wait()
    stdout.join()
    stderr.join()

    if p.returncode != 0 and output_on_error:
        log.stdout("STDOUT:")
        for line in stdout.buffer:
            log.stdout(line)
        log.stderr("STDERR:\r\n")
        for line in stderr.buffer:
            log.stderr(line)

    assert p.returncode == 0, "command failed: {0}".format(cmd.format(*args, **kwargs))
    return "\n".join(stdout.buffer)


class _String(object):
    def __init__(self, s=None):
        self._str = s or ''

    def __enter__(self):
        return self

    def __exit__(self, type, value, tb):
        pass

    def __str__(self):
        return self._str

    def __get__(self):
        return self._str

    def __add__(self, s):
        return self._str + s

    def __iadd__(self, s):
        self._str += s
        return self

    def endswith(self, substr):
        return self._str.endswith(substr)

    def startswith(self, substr):
        return self._str.startswith(substr)


class _tmpdir(object):
    def __init__(self, name, cwd=None):
        self._name = name
        self._path = None
        self._cwd = cwd or os.getcwd()

    def __enter__(self):
        try:
            dirname = self._cwd
            fs.makedirs(dirname)
            self._path = fs.mkdtemp(prefix=self._name + "-", dir=dirname)
        except:
            raise
        assert self._path, "failed to create temporary directory"
        return self

    def __exit__(self, type, value, tb):
        if self._path:
            fs.rmtree(self._path)

    @property
    def path(self):
        return self.get_path()

    def get_path(self):
        return self._path


class _CMake(object):
    def __init__(self, deps, tools):
        self.deps = deps
        self.tools = tools
        self.builddir = self.tools.builddir()
        self.installdir = self.tools.builddir("install")

    def configure(self, sourcedir, *args, **kwargs):
        sourcedir = self.tools.expand_path(sourcedir)

        extra_args = ["-D{0}={1}".format(key, self.tools.expand(val))
                      for key, val in kwargs.items()]
        extra_args = " ".join(extra_args)

        with self.tools.cwd(self.builddir):
            self.tools.run("cmake {0} -DCMAKE_INSTALL_PREFIX={1} {2}",
                           sourcedir, self.installdir, extra_args,
                           output=True)

    def build(self, *args, **kwargs):
        with self.tools.cwd(self.builddir):
            self.tools.run("cmake --build .", output=True)

    def install(self, *args, **kwargs):
        with self.tools.cwd(self.builddir):
            self.tools.run("cmake --build . --target install", output=True)

    def publish(self, artifact, files='*', *args, **kwargs):
        with self.tools.cwd(self.installdir):
            artifact.collect(files, *args, **kwargs)


class _Meson(object):
    def __init__(self, deps, tools):
        self.deps = deps
        self.tools = tools
        self.builddir = self.tools.builddir()
        self.installdir = self.tools.builddir("install")

    def configure(self, sourcedir, *args, **kwargs):
        sourcedir = self.tools.expand_path(sourcedir)
        self.tools.run("meson --prefix=/ {0} {1}", sourcedir, self.builddir,
                       output=True)

    def build(self, *args, **kwargs):
        self.tools.run("ninja -C {0} ", self.builddir, output=True)

    def install(self, *args, **kwargs):
        self.tools.run("DESTDIR={0} ninja -C {1} install",
                       self.installdir, self.builddir,
                       output=True)

    def publish(self, artifact, files='*', *args, **kwargs):
        with self.tools.cwd(self.installdir):
            artifact.collect(files, *args, **kwargs)


class _AutoTools(object):
    def __init__(self, deps, tools):
        self.deps = deps
        self.tools = tools
        self.builddir = self.tools.builddir()
        self.installdir = self.tools.builddir("install")

    def configure(self, sourcedir, *args, **kwargs):
        sourcedir = self.tools.expand_path(sourcedir)

        if not fs.path.exists(fs.path.join(sourcedir, "configure")):
            with self.tools.cwd(sourcedir):
                self.tools.run("autoreconf -visf", output=True)

        with self.tools.cwd(self.builddir):
            self.tools.run("{0}/configure --prefix={1} {2}",
                           sourcedir, self.installdir,
                           self.tools.getenv("CONFIGURE_FLAGS"),
                           output=True)

    def build(self, *args, **kwargs):
        with self.tools.cwd(self.builddir):
            self.tools.run("make VERBOSE=yes Q= V=1 -j{0}",
                           self.tools.cpu_count(), output=True)

    def install(self, *args, **kwargs):
        with self.tools.cwd(self.builddir):
            self.tools.run("make install", output=True)

    def publish(self, artifact, files='*', *args, **kwargs):
        with self.tools.cwd(self.installdir):
            artifact.collect(files, *args, **kwargs)


class Tools(object):
    """ A collection of useful tools.

    Any {keyword} arguments, or macros, found in strings passed to
    tool functions are automatically expanded to the value of the
    associated task's parameters and properties. Relative paths are
    made absolute by prepending the current working directory.
    """

    def __init__(self, task=None, cwd=None):
        self._cwd = cwd or os.getcwd()
        self._env = {key: value for key, value in os.environ.items()}
        self._task = task
        self._builddir = {}

    def __enter__(self):
        return self

    def __exit__(self, type, value, tb):
        for dir in self._builddir.values():
            fs.rmtree(dir)
        return False

    def append_file(self, pathname, content):
        """ Appends data at the end of a file.

        Args:
            pathname (str): Path to file. The file must exist.
            content (str): Data to be appended at the end of the file.
        """

        pathname = self.expand_path(pathname)
        content = self.expand(content)
        with open(path, "ab") as f:
            f.write(content.encode())

    def archive(self, pathname, filename):
        """ Creates a (compressed) archive.

        The type of archive to create is determined by the filename extension.
        Supported formats are:

        - tar
        - tar.bz2
        - tar.gz
        - tar.xz
        - zip

        Args:
            pathname (str): Directory path of files to be archived.
            filename (str): Name/path of created archive.
        """
        filename = self.expand_path(filename)
        pathname = self.expand_path(pathname)

        fmt = None
        if filename.endswith(".zip"):
            fmt = "zip"
            basename = filename[:-4]
        elif filename.endswith(".tar"):
            fmt = "tar"
            basename = filename[:-4]
        elif filename.endswith(".tar.gz"):
            fmt = "gztar"
            basename = filename[:-7]
        elif filename.endswith(".tgz"):
            fmt = "gztar"
            basename = filename[:-4]
        elif filename.endswith(".tar.bz2"):
            fmt = "bztar"
            basename = filename[:-8]
        elif filename.endswith(".tar.xz"):
            fmt = "xztar"
            basename = filename[:-8]
        assert fmt, "unknown filetype '{0}': {1}".format(ext, fs.path.basename(filename))
        try:
            shutil.make_archive(basename, fmt, root_dir=pathname)
            return filename
        except Exception as e:
            assert False, "failed to archive directory: {0}".format(pathname)

    def autotools(self, deps=None):
        """ Creates an AutoTools invokation helper """
        return _AutoTools(deps, self)

    def builddir(self, name="build", incremental=False):
        """ Creates a temporary build directory.

        The build directory will persist for the duration of a task's
        execution. It is automatically removed afterwards.

        Args:
            name (str): Name prefix for the directory. A unique
                autogenerated suffix will also be appended to the
                final name.

        Returns:
            Path to the created directory.
        """
        name = self.expand(name)

        if incremental and self._task is not None:
            unique = utils.sha1(utils.format_task_name(
                self._task.name,
                self._task._get_parameters()))
            dirname = fs.path.join(self.getcwd(), "{0}-{1}-{2}".format(
                name, self._task.canonical_name, unique[:8]))
            fs.makedirs(dirname)
            return dirname

        if name not in self._builddir:
            dirname = self.getcwd()
            fs.makedirs(dirname)
            self._builddir[name] = fs.mkdtemp(prefix=name+"-", dir=dirname)

        return self._builddir[name]


    def chmod(self, pathname, mode):
        """ Changes permissions of files and directories.

        Args:
            pathname (str): Path to a file or directory to change
                permissions for.
            mode (int): Requested permission bits.
        """
        pathname = self.expand_path(pathname)
        return os.chmod(pathname, mode)

    def cmake(self, deps=None):
        """ Creates a CMake invokation helper """
        return _CMake(deps, self)

    def copy(self, src, dst, symlinks=False):
        """ Copies file and directories (recursively).

        The directory tree structure is retained when copying directories.

        Args:
            src (str): Path to a file or directory to be copied.
            dest (str): Destination path. If the string ends with a
                path separator a new directory will
                be created and source files/directories will be copied into
                the new directory. A destination without trailing path
                separator can be used to rename single files, one at a time.
            symlinks (boolean, optional): If True, symlinks are copied.
                The default is False, i.e. the symlink target is copied.
        """
        src = self.expand_path(src)
        dst = self.expand_path(dst)
        return fs.copy(src, dst, symlinks=symlinks)

    def cpu_count(self):
        """ The number of CPUs on the host.

        Returns:
            int: The number of CPUs on the host.
        """

        return multiprocessing.cpu_count()

    @contextmanager
    def cwd(self, pathname, *args, **kwargs):
        """ Change the current working directory to the specified path.

        This function doesn't change the working directory of the Jolt
        process. It only changes the working directory for tools within
        the tools object.

        Args:
            pathname (str): Path to change to.

        Example:

            .. code-block:: python

                with tools.cwd("subdir") as cwd:
                    print(cwd)
        """
        path = self.expand_path(pathname, *args, **kwargs)
        prev = self._cwd
        try:
            assert fs.path.exists(self._cwd) and fs.path.isdir(self._cwd), \
                "failed to change directory to {0}" \
                .format(self._cwd)
            self._cwd = path
            yield fs.path.normpath(self._cwd)
        finally:
            self._cwd = prev

    def download(self, url, pathname, **kwargs):
        """ Downloads a file using HTTP.

        Args:
           url (str): URL to the file to be downloaded.
           pathname (str): Name/path of destination file.
           kwargs (optional): Addidional keyword arguments passed on
               directly ``requests.get()``.

        """
        url = self.expand(url)
        pathname = self.expand_path(pathname)
        try:
            response = requests.get(url, stream=True, **kwargs)
            name = fs.path.basename(pathname)
            size = int(response.headers['content-length'])
            with log.progress("Downloading {0}".format(name), size, "B") as pbar:
                with open(pathname, 'wb') as out_file:
                    chunk_size = 4096
                    for data in response.iter_content(chunk_size=chunk_size):
                        out_file.write(data)
                        pbar.update(len(data))
            return response.status_code == 200
        except:
            log.exception()
            return False

    @contextmanager
    def environ(self, **kwargs):
        """ Set/get environment variables.

        Only child processes spawned by the same tools object will be affected
        by the changed environment.

        The changed environment is only valid within a context and it is
        restored immediately upon leaving the context.

        Args:
            kwargs (optinal): A list of keyword arguments assigning values to
                environment variable.

        Example:

            .. code-block:: python

                with tools.environ(CC="clang"):
                    tools.run("make all")
        """
        for key, value in kwargs.items():
            kwargs[key] = self.expand(value)

        restore = {key: value for key, value in self._env.items()}
        self._env.update(kwargs)
        yield self._env
        for key, value in kwargs.items():
            if key not in restore:
                del self._env[key]
        self._env.update(restore)

    def expand(self, string, *args, **kwargs):
        """ Expands keyword arguments/macros in a format string.

        This function is identical to ``str.format()`` but it
        automatically collects keyword arguments from a task's parameters
        and properties.

        Args:
            string (str): The string to be expanded.
            args (str, optional): Additional positional values required
                by the format string.
            kwargs (str, optional): Additional keyword values required by
                the format string.

        Returns:
            str: Expanded string.

        Example:

            .. code-block:: python

                target = Parameter(default="all")
                verbose = "yes"

                def run(self, deps, tools):
                    print(tools.expand("make {target} VERBOSE={verbose}"))  # "make all VERBOSE=yes"

        """
        return self._task._get_expansion(string, *args, **kwargs) \
            if self._task is not None \
            else utils.expand(string, *args, **kwargs)

    def expand_path(self, pathname, *args, **kwargs):
        """ Expands keyword arguments(macros in a pathname format string.

        This function is identical to ``str.format()`` but it
        automatically collects keyword arguments from a task's parameters
        and properties.

        The function also makes relative paths absolute by prepending the
        current working directory.

        Args:
            pathname (str): The pathname to be expanded.
            args (str, optional): Additional positional values required
                by the format pathname.
            kwargs (str, optional): Additional keyword values required by
                the format pathname.

        Return
            str: Expanded string.
        """

        pathname = fs.path.join(self.getcwd(), pathname)
        return self.expand(pathname, *args, **kwargs)

    def extract(self, filename, pathname, files=None):
        """ Extracts files in an archive.

        Supported formats are:

        - tar
        - tar.bz2
        - tar.gz
        - tar.xz
        - zip

        Args:
            filename (str): Name/path of archive file to be extracted.
            pathname (str): Destination path for extracted files.
            files (list, optional): List of files the be extracted
                from the archive. If not provided, all files are
                extracted.

        """
        filename = self.expand_path(filename)
        filepath = self.expand_path(pathname)
        try:
            fs.makedirs(filepath)
            if filename.endswith(".zip"):
                with zipfile.ZipFile(filename, 'r') as zip:
                    if files:
                        zip.extract(files, filepath)
                    else:
                        zip.extractall(filepath)
            elif filename.endswith(".tar"):
                with tarfile.open(filename, 'r') as tar:
                    if files:
                        tar.extract(files, filepath)
                    else:
                        tar.extractall(filepath)
            elif filename.endswith(".tar.gz") or filename.endswith(".tgz"):
                with tarfile.open(filename, 'r:gz') as tar:
                    if files:
                        tar.extract(files, filepath)
                    else:
                        tar.extractall(filepath)
            elif filename.endswith(".tar.bz2"):
                # bz2file module for multistream support
                with bz2file.open(filename) as bz2:
                    with tarfile.open(fileobj=bz2) as tar:
                        if files:
                            tar.extract(files, filepath)
                        else:
                            tar.extractall(filepath)
            elif filename.endswith(".tar.xz"):
                with tarfile.open(filename, 'r:xz') as tar:
                    if files:
                        tar.extract(files, filepath)
                    else:
                        tar.extractall(filepath)
            else:
                assert False, "unknown filetype: {0}".format(fs.path.basename(filename))
        except Exception as e:
            log.exception()
            assert False, "failed to extract archive: {0}".format(filename)

    def file_size(self, pathname):
        """ Determines the size of a file.

        Args:
            pathname (str): Name/path of file for which the size is requested.

        Returns:
            int: The size of the file in bytes.
        """
        filepath = self.expand_path(pathname)
        try:
            stat = os.stat(pathname)
        except:
            assert False, "file not found: {0}".format(pathname)
        else:
            return stat.st_size

    def getcwd(self):
        """ Returns the current working directory. """
        return fs.path.normpath(self._cwd)

    def getenv(self, key, default=""):
        """ Returns the value of an environment variable.

        Only child processes spawned by the same tools object can see
        the environment variables and their values returned by this method.
        Don't assume the same variables are set in the Jolt process' environment.
        """
        return self._env.get(key, default)

    def glob(self, pathname):
        """ Enumerates files and directories.

        Args:
            pathname (str): A pathname pattern used to match files to be
                included in the returned list of files and directories.
                The pattern may contain simple shell-style
                wildcards such as '*' and '?'. Note: files starting with a
                dot are not matched by these wildcards.

        Returns:
            A list of file and directory pathnames. The pathnames are relative
            to the current working directory unless the ``pathname`` argument
            was absolute.

        Example:

            .. code-block:: python

                textfiles = tools.glob("*.txt")
        """
        path = self.expand_path(pathname)
        files = utils.as_list(glob.glob(path))
        if not fs.path.isabs(pathname):
            files = [file[len(self.getcwd())+1:] for file in files]
        return files

    def map_consecutive(self, callable, iterable):
        """ Same as ``map()``. """
        return utils.map_consecutive(callable, iterable)

    def map_concurrent(self, callable, iterable, max_workers=None):
        """ Concurrent `~map()`.

        Args:
            callable: A callable object to be executed for each item in
                the collection.
            iterable: An iterable collection of items.
            max_workers (optional): The maximum number of worker threads
                allowed to be spawned when performing the work. The
                default is the value returned by
                :func:`jolt.Tools.cpu_count()`.

        Returns:
            list: List of return values.


        Example:

            .. code-block:: python

                def compile(self, srcfile):
                    objfile = srcfile + ".o"
                    tools.run("gcc -c {0} -o {1}", srcfile, objfile)
                    return objfile

                srcfiles = ["test.c", "main.c"]
                objfiles = tools.map_concurrent(compile, srcfiles)

        """
        return utils.map_concurrent(callable, iterable, max_workers)

    def meson(self, deps=None):
        """ Creates a Meson invokation helper """
        return _Meson(deps, self)

    def replace_in_file(self, pathname, search, replace):
        """ Replaces all occurrences of a substring in a file.

        Args:
            pathname (str): Name/path of file to modify.
            search (str): Substring to be replaced.
            replace (str): Replacement substring.

        Example:

            .. code-block:: python

                version = Parameter(default="1.0")

                def run(self, deps, tools):
                    tools.replace_in_file("Makefile", "VERSION := 0.9", "VERSION := {version}")
        """
        pathname = self.expand_path(pathname)
        search = self.expand(search)
        replace = self.expand(replace)
        try:
            with open(pathname) as f:
                data = f.read().decode()
            data = data.replace(search, replace)
            with open(pathname, "wb") as f:
                f.write(data.encode())
        except:
            assert False, "failed to replace string in file: {0}".format(path)

    def rmtree(self, pathname, *args, **kwargs):
        """Removes a directory tree from disk.

        Args:
            pathname (str): Path to the file or directory to be removed.
            ignore_errors (boolean, optional): Ignore files that can't be deleted.
                The default is ``False``.

        """
        cmd = self.expand_path(pathname, *args, **kwargs)
        return fs.rmtree(pathname, **kwargs)

    def run(self, cmd, *args, **kwargs):
        """ Runs a command in a shell interpreter.

        Args:
            cmd (str): Command format string.
            args: Positional arguments for the command format string.
            kwargs: Keyword arguments for the command format string.
            output (boolean, optional): By default, the executed command's
                output will be written to the console. Set to ``False`` to
                disable all output.
            output_on_error (boolean, optional): If ``True``, no output is
                written to the console unless the command fails.
                The default is ``False``.

        Example:

            .. code-block:: python

                target = Parameter(default="all")
                verbose = "yes"

                def run(self, deps, tools):
                    tools.run("make {target} VERBOSE={verbose} JOBS={0}", tools.cpu_count())

        """
        cmd = self.expand(cmd, *args, **kwargs)
        stdi, stdo, stde = None, None, None
        try:
            stdi, stdo, stde = None, None, None
            try:
                stdi = termios.tcgetattr(sys.stdin.fileno())
                stdo = termios.tcgetattr(sys.stdout.fileno())
                stde = termios.tcgetattr(sys.stderr.fileno())
            except:
                pass
            return _run(cmd, self._cwd, self._env, *args, **kwargs)
        finally:
            if stdi:
                termios.tcsetattr(sys.stdin.fileno(), termios.TCSANOW, stdi)
            if stdo:
                termios.tcsetattr(sys.stdout.fileno(), termios.TCSANOW, stdo)
            if stde:
                termios.tcsetattr(sys.stderr.fileno(), termios.TCSANOW, stde)

    def setenv(self, key, value=None):
        """ Sets or unset an environment variable.

        Only child processes spawned by the same tools object can see
        the set environment variable and its value.
        Don't assume the same variable is set in the Jolt process' environment.

        Args:
            key (str): Name of variable to set.
            value (str): Value of the variable or ``None`` to unset it.

        """
        if value is None:
            try:
                del self._env[key]
            except:
                pass
        else:
            self._env[key] = self.expand(value)

    def symlink(self, src, dst):
        """ Creates a symbolic link.

        Args:
            src (str): Path to target file or directory.
            dst (str): Name/path of symbolic link.
        """
        src = self.expand_path(src)
        dst = self.expand_path(dst)
        fs.symlink(src, dst)

    def tmpdir(self, name):
        """ Creates a temporary directory.

        The directory is only valid within a context and it is removed
        immediately upon leaving the context.

        Args:
            name (str): Name prefix for the directory. A unique
                autogenerated suffix will also be appended to the
                final name.

        Example:

            .. code-block:: python

                with tools.tmpdir("temp") as tmp, tools.cwd(tmp.path):
                    tools.write_file("tempfile", "tempdata")

        """
        return _tmpdir(name, cwd=self._cwd)

    def unlink(self, pathname, *args, **kwargs):
        """Removes a file from disk.

        To remove directories, use :func:`~jolt.Tools.rmtree`.

        Args:
            pathname (str): Path to the file to be removed.

        """
        cmd = self.expand_path(pathname, *args, **kwargs)
        return fs.unlink(pathname)

    def upload(self, pathname, url, auth=None, **kwargs):
        """ Uploads a file using HTTP (PUT).

        Args:
           pathname (str): Name/path of file to be uploaded.
           url (str): Destination URL.
           auth (requests.auth.AuthBase, optional): Authentication helper.
               See requests.auth for details.
           kwargs (optional): Addidional keyword arguments passed on
               directly to `~requests.put()`.

        """

        pathname = self.expand_path(pathname)
        try:
            name = fs.path.basename(pathname)
            size = self.file_size(pathname)
            with log.progress("Uploading " + name, size, "B") as pbar, \
                 open(pathname, 'rb') as fileobj:
                def read():
                    data = fileobj.read(4096)
                    pbar.update(len(data))
                    return data
                response = requests.put(url, data=iter(read, b''), auth=auth, **kwargs)
                return response.status_code == 201
        except:
            log.exception()
            pass
        return False

    def write_file(self, pathname, content=None):
        """ Creates a file.

        Note:
            Existing files are overwritten.

        Args:
            pathname (str): Name/path of file to be created.
            content (str, optional): Data to be written to the file.
        """
        pathname = self.expand_path(pathname)
        content = self.expand(content) if content is not None else ''
        with open(pathname, "wb") as f:
            f.write(content.encode())



import platform
from jolt.tasks import *
from jolt import influence
from jolt import utils

import ninja_syntax as ninja


class Variable(object):
    def __init__(self, value=None):
        self._value = None

    def create(self, project, writer, deps, tools):
        writer.variable(self.name, self._value)


class EnvironmentVariable(Variable):
    def __init__(self, default=None):
        self._default = default or ''

    def create(self, project, writer, deps, tools):
        writer.variable(self.name, tools.getenv(self.name.upper(), self._default))


class ProjectVariable(Variable):
    def __init__(self, default=None):
        self._default = default or ''

    def create(self, project, writer, deps, tools):
        writer.variable(self.name, getattr(project, self.name))


class Rule(object):
    def __init__(self, command, variables=None, depfile=None, suffix=None,
                 prefix=None, files=None, ):
        self.command = command
        self.variables = variables
        self.depfile = depfile
        self.files = files or []
        self.prefix = prefix or ''
        self.suffix = suffix or ''

    def outfile(self, project, infile):
        dirname, basename = fs.path.split(infile)
        basename = self.prefix + basename + self.suffix
        outfile = fs.path.join(dirname, basename)
        if outfile.startswith(project.joltdir):
            outfile = outfile[len(project.joltdir)+1:]
            outfile = fs.path.join(project.outdir, outfile)
        return outfile

    def create(self, project, writer, deps, tools):
        if self.command is not None:
            writer.rule(self.name, tools.expand(self.command), depfile=self.depfile)
            writer.newline()

    def build(self, project, writer, infiles):
        outfiles = []
        for infile in utils.as_list(infiles):
            outfile = self.outfile(project, infile)
            writer.build(outfile, self.name, infile)
            outfiles.append(outfile)
        return outfiles


class Skip(Rule):
    def __init__(self, files=None):
        self.files = files
        self.command = None

    def create(self, project, writer, deps, tools):
        pass

    def build(self, project, writer, infiles):
        return None


class Objects(Rule):
    def __init__(self, files=None):
        self.files = files
        self.command = None

    def create(self, project, writer, deps, tools):
        pass

    def build(self, project, writer, infiles):
        project.objects.extend(utils.as_list(infiles))
        return None


class GNUCompiler(Rule):
    pass


class FileListWriter(Rule):
    def __init__(self, name):
        self.name = name

    def _write(self, flp, flhp, data, digest):
        with open(flp, "w") as f:
            f.write(data)
        with open(flhp, "w") as f:
            f.write(digest)

    def _identical(self, flp, flhp, data, digest):
        if not fs.path.exists(flp) or not fs.path.exists(flhp):
            return False

        try:
            with open(flhp, "r") as f:
                disk_digest = f.read()
        except:
            return False

        return digest == disk_digest

    def _data(self, files):
        data = "\n".join(files)
        return data, utils.sha1(data)

    def build(self, project, writer, infiles):
        file_list_path = fs.path.join(project.outdir, "{0}.list".format(self.name))
        file_list_hash_path = fs.path.join(project.outdir, "{0}.hash".format(self.name))
        data, digest = self._data(infiles)
        if not self._identical(file_list_path, file_list_hash_path, data, digest):
            self._write(file_list_path, file_list_hash_path, data, digest)
        project.depimports.append(file_list_path)


class GNULinker(Rule):
    def build(self, project, writer, infiles):
        file_list = FileListWriter("objects")
        file_list.build(project, writer, infiles)

        outfile = self.outfile(project, project.binary)
        writer.build(outfile, self.name, infiles, implicit=project.depimports)
        return outfile


class GNUArchiver(Rule):
    def build(self, project, writer, infiles):
        file_list = FileListWriter("objects")
        file_list.build(project, writer, infiles)

        outfile = self.outfile(project, project.binary)
        writer.build(outfile, self.name, infiles, implicit=project.depimports)
        return outfile


class GNUDepImporter(Rule):
    def __init__(self, prefix=None, suffix=None):
        self.prefix = prefix
        self.suffix = suffix
        self.files = []
        self.command = None

    def _build_archives(self, project, writer, deps):
        archives = []
        for name, artifact in deps.items():
             for lib in artifact.cxxinfo.libraries.items():
                 name = "{0}{1}{2}".format(self.prefix, lib, self.suffix)
                 for path in artifact.cxxinfo.libpaths.items():
                     archive = fs.path.join(artifact.stable_path, path, name)
                     if fs.path.exists(archive):
                         archives.append(archive)
        return archives

    def build(self, project, writer, deps):
        imports = []
        if isinstance(project, CXXExecutable):
            imports += self._build_archives(project, writer, deps)
        return imports



class Toolchain(object):
    def __init__(self):
        self._build_variables()
        self._rule_map = self.build_rule_map(self)

    @staticmethod
    def build_rule_map(cls):
        rule_map = {}
        for name, rule in Toolchain.all_rules(cls):
            rule.name = name
            for ext in rule.files:
                rule_map[ext] = rule
        return rule_map

    def _build_variables(self):
        for name, var in Toolchain.all_variables(self):
            var.name = name

    @staticmethod
    def all_rules(cls):
        return [(key, getattr(cls, key)) for key in dir(cls)
                if isinstance(utils.getattr_safe(cls, key), Rule)]

    def find_rule(self, ext):
        rule = self._rule_map.get(ext)
        assert rule, "no build rule match for file with extension '{0}'".format(ext)
        return rule

    @staticmethod
    def all_variables(cls):
        return [(key, getattr(cls, key)) for key in dir(cls)
                if isinstance(utils.getattr_safe(cls, key), Variable)]


class Macros(Variable):
    def __init__(self, prefix=None):
        self.prefix = prefix or ''

    def create(self, project, writer, deps, tools):
        macros = [tools.expand(macro) for macro in project.macros]
        for name, artifact in deps.items():
            macros += artifact.cxxinfo.macros.items()
        macros = ["{0}{1}".format(self.prefix, macro) for macro in macros]
        writer.variable(self.name, " ".join(macros))


class IncludePaths(Variable):
    def __init__(self, prefix=None):
        self.prefix = prefix or ''

    def create(self, project, writer, deps, tools):
        incpaths = [tools.expand_path(path) for path in project.incpaths]
        for name, artifact in deps.items():
            incpaths += [fs.path.join(artifact.stable_path, path)
                         for path in artifact.cxxinfo.incpaths.items()]
        incpaths = ["{0}{1}".format(self.prefix, path) for path in incpaths]
        writer.variable(self.name, " ".join(incpaths))


class LibraryPaths(Variable):
    def __init__(self, prefix=None):
        self.prefix = prefix or ''

    def create(self, project, writer, deps, tools):
        if isinstance(project, CXXLibrary):
            return
        libpaths = [tools.expand_path(path) for path in project.libpaths]
        for name, artifact in deps.items():
            libpaths += [fs.path.join(artifact.stable_path, path)
                         for path in artifact.cxxinfo.libpaths.items()]
        libpaths = ["{0}{1}".format(self.prefix, path) for path in libpaths]
        writer.variable(self.name, " ".join(libpaths))


class Libraries(Variable):
    def __init__(self, prefix=None):
        self.prefix = prefix or ''

    def create(self, project, writer, deps, tools):
        if isinstance(project, CXXLibrary):
            return
        libraries = [tools.expand(lib) for lib in project.libraries]
        for name, artifact in deps.items():
            libraries += artifact.cxxinfo.libraries.items()
        libraries = ["{0}{1}".format(self.prefix, path) for path in libraries]
        writer.variable(self.name, " ".join(libraries))


class GNUToolchain(Toolchain):
    hh = Skip(files=[".h", ".hh", ".hpp", ".hxx"])
    obj = Objects(files=[".o", ".obj", ".a"])

    outdir = ProjectVariable()

    ar = EnvironmentVariable(default="ar")
    cc = EnvironmentVariable(default="gcc")
    cxx = EnvironmentVariable(default="g++")
    ld = EnvironmentVariable(default="g++")

    asflags = EnvironmentVariable(default="")
    cflags = EnvironmentVariable(default="")
    cxxflags = EnvironmentVariable(default="")
    ldflags = EnvironmentVariable(default="")

    macros = Macros(prefix="-D")
    incpaths = IncludePaths(prefix="-I")
    libpaths = LibraryPaths(prefix="-L")
    libraries = Libraries(prefix="-l")

    compile_c = GNUCompiler(
        command="$cc -x c $cflags $extra_cflags $macros $incpaths -MMD -MF $out.d -c $in -o $out",
        depfile="$out.d",
        files=[".c"],
        suffix=".o")

    compile_cxx = GNUCompiler(
        command="$cxx -x c++ $cxxflags $extra_cxxflags $macros $incpaths -MMD -MF $out.d -c $in -o $out",
        depfile="$out.d",
        files=[".cc", ".cpp", ".cxx"],
        suffix=".o")

    compile_asm = GNUCompiler(
        command="$cc -x assembler $asflags $extra_asflags -MMD -MF $out.d -c $in -o $out",
        depfile="$out.d",
        files=[".s", ".asm"],
        suffix=".o")

    compile_asm_with_cpp = GNUCompiler(
        "$cc -x assembler-with-cpp $cflags $extra_cflags $macros $incpaths -MMD -MF $out.d -c $in -o $out",
        depfile="$out.d",
        files=[".S"],
        suffix=".o")

    link = GNULinker(
        command="$ld $ldflags $extra_ldflags $libpaths -Wl,--start-group @objects.list -Wl,--end-group -o $out $libraries")

    archive = GNUArchiver(
        command="$ar cr $out @objects.list",
        prefix="lib",
        suffix=".a")

    depimport = GNUDepImporter(
        prefix="lib",
        suffix=".a")


toolchain = GNUToolchain()


@influence.attribute("incpaths")
@influence.attribute("macros")
@influence.attribute("sources")
@influence.attribute("binary")
class CXXProject(Task):
    incpaths = []
    macros = []
    sources = []
    depimports = []
    source_influence = True
    binary = None
    incremental = True

    def __init__(self, *args, **kwargs):
        super(CXXProject, self).__init__(*args, **kwargs)
        self._init_sources()
        self.macros = utils.as_list(utils.call_or_return(self, self.__class__.macros))
        self.incpaths = utils.as_list(utils.call_or_return(self, self.__class__.incpaths))
        self.binary = self.__class__.binary or self.canonical_name
        if self.source_influence:
            for source in self.sources:
                self.influence.append(influence.FileInfluence(source))
        self._rule_map = Toolchain.build_rule_map(self)

    def _init_sources(self):
        sources = utils.as_list(utils.call_or_return(self, self.__class__.sources))
        self.sources = []
        for l in map(self.tools.glob, sources):
            self.sources += l
        self.sources.sort()
        assert self.sources, "no source files found for task {0}".format(self.name)

    def _write_ninja_file(self, basedir, deps, tools):
        with open(fs.path.join(basedir, "build.ninja"), "w") as fobj:
            writer = ninja.Writer(fobj)
            self._populate_variables(writer, deps, tools)
            self._populate_rules(writer, deps, tools)
            self._populate_inputs(writer, deps, tools)
            self._populate_project(writer, deps, tools)
            writer.close()

    def find_rule(self, ext):
        rule = self._rule_map.get(ext)
        if rule is None:
            rule = toolchain.find_rule(ext)
        assert rule, "no build rule match for file with extension '{0}'".format(ext)
        return rule

    def _populate_variables(self, writer, deps, tools):
        for name, var in Toolchain.all_variables(toolchain):
            var.create(self, writer, deps, tools)
        for name, var in Toolchain.all_variables(self):
            var.create(self, writer, deps, tools)
        writer.newline()

    def _populate_rules(self, writer, deps, tools):
        for name, rule in Toolchain.all_rules(toolchain):
            rule.create(self, writer, deps, tools)
        for name, rule in Toolchain.all_rules(self):
            rule.create(self, writer, deps, tools)
        writer.newline()

    def _populate_inputs(self, writer, deps, tools):
        self.objects = []
        sources = copy.copy(self.sources)
        while sources:
            source = sources.pop()
            _, ext = fs.path.splitext(source)
            rule = self.find_rule(ext)
            output = rule.build(self, writer, tools.expand_path(source))
            sources.extend(output or [])
        writer.newline()

    def _populate_project(self, writer, deps, tools):
        pass

    def run(self, deps, tools):
        self.outdir = tools.builddir("build/ninja", self.incremental)
        self._write_ninja_file(self.outdir, deps, tools)
        tools.run("ninja -C {0}", self.outdir)


@influence.attribute("shared")
class CXXLibrary(CXXProject):
    shared = False

    def __init__(self, *args, **kwargs):
        super(CXXLibrary, self).__init__(*args, **kwargs)

    def _populate_inputs(self, writer, deps, tools):
        self.depimports += toolchain.depimport.build(self, writer, deps)
        super(CXXLibrary, self)._populate_inputs(writer, deps, tools)

    def _populate_project(self, writer, deps, tools):
        toolchain.archive.build(self, writer, self.objects)

    def publish(self, artifact, tools):
        with tools.cwd(self.outdir):
            artifact.collect("*.a", "lib/")
            artifact.collect("*.so", "lib/")
            artifact.collect("*.dll", "lib/")
        artifact.cxxinfo.libpaths.append("lib")
        artifact.cxxinfo.libraries.append(self.binary)


@influence.attribute("libpaths")
@influence.attribute("libraries")
class CXXExecutable(CXXProject):
    libpaths = []
    libraries = []

    def __init__(self, *args, **kwargs):
        super(CXXExecutable, self).__init__(*args, **kwargs)
        self.libpaths = utils.as_list(utils.call_or_return(self, self.__class__.libpaths))
        self.libraries = utils.as_list(utils.call_or_return(self, self.__class__.libraries))

    def _populate_inputs(self, writer, deps, tools):
        self.depimports += toolchain.depimport.build(self, writer, deps)
        super(CXXExecutable, self)._populate_inputs(writer, deps, tools)

    def _populate_project(self, writer, deps, tools):
        toolchain.link.build(self, writer, self.objects)

    def publish(self, artifact, tools):
        with tools.cwd(self.outdir):
            if platform.system() == "Windows":
                artifact.collect(self.binary + '.exe', "bin/")
            else:
                artifact.collect(self.binary, "bin/")
        artifact.environ.PATH.append("bin")



##############################################################################
#
# (C) 2016 - Robert Andersson - All rights reserved.
#
# This file and its contents are the property of Robert Andersson
# and may not be distributed, copied, or disclosed, in whole or in part,
# for any reason without written consent of the copyright holder.
#
##############################################################################


from .tools import Tool
from .tools.directory import PyBuildDirectoryCreator
from .projects import pybuild
from utils import DepfileParser
from build.model import CXXLibrary
from os import path, pathsep, environ
from copy import copy
from functools import partial


class _ExecutableMixin(object):
    def __init__(self, executable, filetype=None, prefix=None, path=None, *args, **kwargs):
        super(_ExecutableMixin, self).__init__(*args, **kwargs)
        self.prefix = prefix or ''
        self.bare_executable = executable
        self.executable = self.prefix + self.bare_executable
        self.filetype = filetype
        self.path = path or ''
        self.environ = copy(environ)
        if self.path and self.environ["PATH"]:
            self.environ["PATH"] += pathsep + path


def _scan_deps(cxx_project, obj):
    depfile, _ = path.splitext(obj.product)
    obj.clear_hash()
    for dep in DepfileParser(depfile + ".d").dependencies:
        cxx_project.add_source(dep)
        cxx_project.add_dependency(obj.product, dep)


class PyBuildCXXCompiler(_ExecutableMixin, Tool):
    def __init__(self, executable='g++', filetype='c++', flags=None, *args, **kwargs):
        super(PyBuildCXXCompiler, self).__init__(executable=executable, filetype=filetype, *args, **kwargs)
        self._output_ext = '.o'
        self._flags = flags or ''

    def _directory(self, cxx_project, dirname):
        return PyBuildDirectoryCreator().transform(cxx_project, dirname)

    def _product(self, cxx_project, source_file):
        return '{output}/{}{}'.format(source_file, self._output_ext, output=cxx_project.output)

    def _cmdline(self, cxx_project, source_file):
        flags = cxx_project.cflags if self.filetype != 'c++' else cxx_project.cxxflags

        return "{} -x {} {} -MMD -c {} -o {}".format(
            self.executable,
            self.filetype,
            ' '.join(flags + [self._flags]),
            source_file,
            self._product(cxx_project, source_file))

    def _info(self, source_file):
        return ' [{}] {}'.format(self.bare_executable.upper(), source_file)

    def transform(self, cxx_project, source_file):
        product = self._product(cxx_project, source_file.path)
        dir = self._directory(cxx_project, path.dirname(product))
        obj =  pybuild.Object(
            product,
            self._cmdline(cxx_project, source_file.path),
            self._info(source_file.path),
            self.environ)
        cxx_project.add_source(source_file.path)
        cxx_project.add_job(obj)
        cxx_project.add_dependency(obj.product, source_file.path)
        cxx_project.add_dependency(obj.product, dir.product)

        obj.on_completed = partial(_scan_deps, cxx_project)
        obj.on_completed(obj)

        return obj


class PyBuildCXXArchiver(_ExecutableMixin, Tool):
    def __init__(self, executable='ar', *args, **kwargs):
        super(PyBuildCXXArchiver, self).__init__(executable=executable, *args, **kwargs)
        self._output_pfx = 'lib'
        self._output_ext = '.a'

    def _directory(self, cxx_project, dirname):
        return PyBuildDirectoryCreator().transform(cxx_project, dirname)

    def _product(self, cxx_project):
        return '{output}/{}{}{}'.format(self._output_pfx, cxx_project.name, self._output_ext, output=cxx_project.output)

    def _filelist(self, cxx_project):
        return self._product(cxx_project) + ".objects"

    def _cmdline(self, cxx_project, object_files):
        return "{} cr {} @{}".format(self.executable, self._product(cxx_project), self._filelist(cxx_project))

    def _info(self, cxx_project):
        return ' [{}] {}'.format(self.bare_executable.upper(), cxx_project.name)

    def transform(self, project, cxx_project, object_files):
        product = self._product(cxx_project)
        dir = self._directory(cxx_project, path.dirname(product))
        library = pybuild.Object(
            product,
            self._cmdline(cxx_project, object_files),
            self._info(cxx_project),
            self.environ)
        filelist = cxx_project.add_filelist(self._filelist(cxx_project), object_files)
        cxx_project.add_job(library)
        cxx_project.add_dependency(library.product, dir.product)
        cxx_project.add_dependency(library.product, filelist.product)
        cxx_project.add_dependency(filelist.product, dir.product)
        for obj in object_files:
            cxx_project.add_dependency(library.product, obj)

        library.on_completed = partial(_scan_deps, cxx_project)
        library.on_completed(library)

        return library


class PyBuildCXXLinker(_ExecutableMixin, Tool):
    def __init__(self, executable='g++', flags=None, *args, **kwargs):
        super(PyBuildCXXLinker, self).__init__(executable=executable, *args, **kwargs)
        self._output_ext = ''
        self._output_pfx_shared = 'lib'
        self._output_ext_shared = '.so'
        self._flags = flags or ''

    def _directory(self, cxx_project, dirname):
        return PyBuildDirectoryCreator().transform(cxx_project, dirname)

    def _product(self, cxx_project):
        if isinstance(cxx_project.project, CXXLibrary) and cxx_project.project.shared:
            return '{output}/{}{}{}'.format(
                self._output_pfx_shared,
                cxx_project.name,
                self._output_ext_shared,
                output=cxx_project.output)
        return '{output}/{}{}'.format(cxx_project.name, self._output_ext, output=cxx_project.output)


    def _filelist(self, cxx_project):
        return self._product(cxx_project) + ".objects"

    def _cmdline(self, project, cxx_project, object_files):
        libraries = ['-l{}'.format(path) for path in cxx_project.libraries]
        flags = cxx_project.linkflags

        return "{} {} @{} -o {} -Wl,--start-group {} -Wl,--end-group".format(
            self.executable,
            ' '.join(flags + [self._flags]),
            self._filelist(cxx_project),
            self._product(cxx_project),
            ' '.join(libraries))

    def _info(self, cxx_project):
        return ' [{}] {}'.format(self.bare_executable.upper(), cxx_project.name)

    def transform(self, project, cxx_project, object_files):
        product = self._product(cxx_project)
        dir = self._directory(cxx_project, path.dirname(product))
        executable = pybuild.Object(
            product,
            self._cmdline(project, cxx_project, object_files),
            self._info(cxx_project),
            self.environ)
        filelist = cxx_project.add_filelist(self._filelist(cxx_project), object_files)
        cxx_project.add_job(executable)
        cxx_project.add_dependency(executable.product, dir.product)
        cxx_project.add_dependency(executable.product, filelist.product)
        cxx_project.add_dependency(filelist.product, dir.product)
        for obj in object_files:
            cxx_project.add_dependency(executable.product, obj)

        executable.on_completed = partial(_scan_deps, cxx_project)
        executable.on_completed(executable)

        return executable


class GNUToolFactory:
    def __init__(self, prefix=None, path=None, env=None):
        self.prefix = prefix
        self.path = path
        self.env = env or environ
        self.CC = self.env.get("CC", "gcc")
        self.CXX = self.env.get("CXX", "g++")
        self.AR = self.env.get("AR", "ar")
        self.AS = self.env.get("AS", "gcc")
        self.LD = self.env.get("LD", "g++")
        self.asflags = self.env.get("ASFLAGS", "")
        self.cflags = self.env.get("CFLAGS", "")
        self.cxxflags = self.env.get("CXXFLAGS", "")
        self.ldflags = self.env.get("LDFLAGS", "")

    def configure(self, toolchain):
        toolchain.add_tool('.s', PyBuildCXXCompiler(self.AS, 'assembler', flags=self.asflags, prefix=self.prefix, path=self.path))
        toolchain.add_tool('.S', PyBuildCXXCompiler(self.AS, 'assembler-with-cpp', flags=self.asflags, prefix=self.prefix, path=self.path))
        toolchain.add_tool('.c', PyBuildCXXCompiler(self.CC, 'c', flags=self.cflags, prefix=self.prefix, path=self.path))
        toolchain.add_tool('.cc', PyBuildCXXCompiler(self.CXX, 'c++', flags=self.cflags, prefix=self.prefix, path=self.path))
        toolchain.add_tool('.cpp', PyBuildCXXCompiler(self.CXX, 'c++', flags=self.cxxflags, prefix=self.prefix, path=self.path))
        toolchain.add_tool('.cxx', PyBuildCXXCompiler(self.CXX, 'c++', flags=self.cxxflags, prefix=self.prefix, path=self.path))
        toolchain.archiver = PyBuildCXXArchiver(self.AR, prefix=self.prefix, path=self.path)
        toolchain.linker = PyBuildCXXLinker(executable=self.LD, flags=self.ldflags, prefix=self.prefix, path=self.path)



import click
import imp
import traceback
import subprocess
import signal
import sys
import webbrowser


from jolt.tasks import Task, TaskRegistry, Parameter
from jolt import scheduler
from jolt import graph
from jolt import cache
from jolt import filesystem as fs
from jolt import log
from jolt.log import path as log_path
from jolt import config
from jolt import plugins
from jolt.plugins import cxxinfo, environ, strings
from jolt import loader
from jolt import utils
from jolt.influence import *
from jolt.options import JoltOptions


@click.group()
@click.option("-v", "--verbose", is_flag=True, help="Verbose.")
@click.option("-vv", "--extra-verbose", is_flag=True, help="Verbose.")
@click.option("-c", "--config-file", type=str, help="Configuration file")
def cli(verbose, extra_verbose, config_file):
    """ Jolt - a task execution tool. """

    if verbose:
        log.set_level(log.VERBOSE)
    if extra_verbose:
        log.set_level(log.HYSTERICAL)
    if config_file:
        config.load(config_file)

    # Load configured plugins
    for section in config.sections():
        path = fs.path.dirname(__file__)
        path = fs.path.join(path, "plugins", section + ".py")
        if fs.path.exists(path):
            imp.load_source("jolt.plugins." + section, path)

    tasks, tests = loader.JoltLoader().get().load()
    for cls in tasks:
        TaskRegistry.get().add_task_class(cls)
    for cls in tests:
        TaskRegistry.get().add_test_class(cls)


def _autocomplete_tasks(ctx, args, incomplete):
    tasks, tests = loader.JoltLoader.get().load()
    tasks = [task.name for task in tasks + tests if task.name.startswith(incomplete or '')]
    return sorted(tasks)


@cli.command()
@click.argument("task", type=str, nargs=-1, required=True, autocompletion=_autocomplete_tasks)
@click.option("-n", "--network", is_flag=True, default=False, help="Build on network.")
@click.option("-l", "--local", is_flag=True, default=False, help="Disable all network operations.")
@click.option("-k", "--keep-going", is_flag=True, default=False, help="Build as many tasks as possible, don't abort on first failure.")
@click.option("-i", "--identity", type=str, help="Expected hash identity")
@click.option("--no-download", is_flag=True, default=False,
              help="Don't download artifacts from remote storage")
@click.option("--no-upload", is_flag=True, default=False,
              help="Don't upload artifacts to remote storage")
@click.option("--download", is_flag=True, default=False,
              help="Do download artifacts from remote storage")
@click.option("--upload", is_flag=True, default=False,
              help="Do upload artifacts to remote storage")
@click.option("--worker", is_flag=True, default=False,
              help="Run with the worker build strategy", hidden=True)
@click.option("-d", "--default", type=str, multiple=True, help="Override default parameter values.")
def build(task, network, keep_going, identity, default, local,
          no_download, no_upload, download, upload, worker):
    """
    Execute specified task.

    <WIP>
    """
    if network:
        download = config.getboolean("network", "download", True)
        upload = config.getboolean("network", "upload", True)
    else:
        download = config.getboolean("jolt", "download", True)
        upload = config.getboolean("jolt", "upload", True)

    if no_download:
        download = False
    if no_upload:
        upload = False
    if download:
        download = True
    if upload:
        upload = True

    if local:
        download = False
        upload = False

    options = JoltOptions(network=network,
                          download=download,
                          upload=upload,
                          keep_going=keep_going,
                          default=default)

    acache = cache.ArtifactCache.get(options)

    executors = scheduler.ExecutorRegistry.get(options)
    if worker:
        log.verbose("Local build as a worker")
        strategy = scheduler.WorkerStrategy(executors, acache)
    elif network:
        log.verbose("Distributed build as a user")
        strategy = scheduler.DistributedStrategy(executors, acache)
    else:
        log.verbose("Local build as a user")
        strategy = scheduler.LocalStrategy(executors, acache)

    registry = TaskRegistry.get()

    for params in default:
        registry.set_default_parameters(params)

    gb = graph.GraphBuilder(registry)
    dag = gb.build(task)

    # Inform cache about what task artifacts we will need.
    acache.advise(dag.tasks)

    if identity:
        root = dag.select(lambda graph, task: task.identity.startswith(identity))
        assert len(root) >= 1, "unknown hash identity, no such task: {0}".format(identity)

    queue = scheduler.TaskQueue(strategy)

    try:
        while dag.has_tasks():
            leafs = dag.select(lambda graph, task: task.is_ready())
            while leafs:
                task = leafs.pop()
                queue.submit(acache, task)

            task, error = queue.wait()
            if not task:
                dag.debug()
                assert task, "no more tasks in progress, only blocked tasks remain"

            if not keep_going and error is not None:
                queue.abort()
                raise error
    except KeyboardInterrupt:
        log.warn("Interrupted by user")
        try:
            queue.abort()
            return
        except KeyboardInterrupt:
            log.warn("Interrupted again, exiting")
            os._exit(1)


@cli.command()
@click.argument("task", type=str, nargs=-1, required=False, autocompletion=_autocomplete_tasks)
def clean(task):
    """
    Remove (task artifact from) local cache.

    <WIP>
    """
    acache = cache.ArtifactCache.get()
    if task:
        registry = TaskRegistry.get()
        dag = graph.GraphBuilder(registry).build(task)
        tasks = dag.select(
            lambda graph, node: node.short_qualified_name in task or \
            node.qualified_name in task)
        for task in tasks:
            acache.discard(task)
    else:
        fs.rmtree(acache.root)


@cli.command()
@click.argument("task", type=str, nargs=-1, required=False, autocompletion=_autocomplete_tasks)
@click.option("-p", "--prune", is_flag=True, help="Omit tasks with cached artifacts.")
def display(task, prune):
    """
    Display a task and its dependencies visually.

    <WIP>
    """
    options = JoltOptions()
    registry = TaskRegistry.get()
    gb = graph.GraphBuilder(registry)
    dag = gb.build(task)
    if prune:
        acache = cache.ArtifactCache.get()
        dag.prune(lambda graph, task: task.is_available_locally(acache) or task.is_resource())
    if dag.has_tasks():
        try:
            gb.display()
        except Exception as e:
            if "requires pygraphviz" in str(e):
                assert False, "this features requires pygraphviz, please install it"
            assert False, "an exception occurred during task dependency evaluation, see log for details"
    else:
        log.info("no tasks to display")


@cli.command()
@click.option("-f", "--follow", is_flag=True, help="Display log output as it appears")
@click.option("-D", "--delete", is_flag=True, help="Delete the log file")
def docs(follow, delete):
    """
    Opens the Jolt documentation in the default webbrowser.
    """
    webbrowser.open("http://jolt.readthedocs.io/")


@cli.command(name="list")
@click.argument("task", type=str, nargs=-1, required=False, autocompletion=_autocomplete_tasks)
@click.option("-a", "--all", is_flag=True, help="Print all tasks recursively")
def _list(task=None, reverse=False, all=False):
    """
    List all tasks, or dependencies of a specific task.

    <WIP>
    """

    registry = TaskRegistry.get()

    if not task:
        classes = registry.get_task_classes()
        classes += registry.get_test_classes()
        for task in sorted(classes, key=lambda x: x.name):
            if task.name:
                print(task.name)
        return

    try:
        dag = graph.GraphBuilder(registry).build(task)
    except:
        assert False, "an exception occurred during task dependency evaluation, see log for details"

    tasks = dag.select(lambda graph, node: \
                       node.short_qualified_name in task or \
                       node.qualified_name in task)
    successors = set()
    for task in tasks:
        for successor in dag.successors(task):
            successors.add(successor.qualified_name)

    for task in sorted(list(successors)):
        print(task)


@cli.command(name="log")
@click.option("-f", "--follow", is_flag=True, help="Display log output as it appears")
@click.option("-D", "--delete", is_flag=True, help="Delete the log file")
def _log(follow, delete):
    """
    Access the Jolt log file.

    <WIP>
    """
    if follow:
        subprocess.call("tail -f {0}".format(log_path), shell=True)
    elif delete:
        fs.unlink(log_path)
    else:
        subprocess.call("less {0}".format(log_path), shell=True)


@cli.command()
@click.argument("task", autocompletion=_autocomplete_tasks)
@click.option("-i", "--influence", is_flag=True, help="Print task influence.")
@click.option("-a", "--artifacts", is_flag=True, help="Print task artifact status.")
def info(task, influence=False, artifacts=False):
    """
    View information about a task, including its documentation.

    <WIP>
    """
    task_name = task
    task_cls_name, task_params = utils.parse_task_name(task_name)
    task_registry = TaskRegistry.get()
    task = task_registry.get_task_class(task_cls_name)
    assert task, "no such task: {0}".format(task_name)

    click.echo()
    click.echo("  {0}".format(task.name))
    click.echo()
    if task.__doc__:
        click.echo("  {0}".format(task.__doc__.strip()))
        click.echo()
    click.echo("  Parameters")
    has_param = False
    for item, param in task.__dict__.items():
        if isinstance(param, Parameter):
            has_param = True
            click.echo("    {0:<15}   {1}".format(item, param.__doc__ or ""))
    if not has_param:
        click.echo("    None")

    click.echo()
    click.echo("  Requirements")
    try:
        task = task_registry.get_task(task_name)
        for req in utils.as_list(utils.call_or_return(task, task.requires)):
            click.echo("    {0}".format(task.tools.expand(req)))
        if not task.requires:
            click.echo("    None")
        click.echo()
    except Exception as e:
        if "has not been set" in str(e):
            click.echo("    Unavailable (parameters must be set)")
            click.echo()
            return
        click.echo("    Unavailable (exception during evaluation)")
        click.echo()


    if artifacts:
        acache = cache.ArtifactCache.get()
        dag = graph.GraphBuilder(task_registry).build(
            [utils.format_task_name(task.name, task._get_parameters())])
        tasks = dag.select(lambda graph, node: graph.is_root(node))
        assert len(tasks) == 1, "unexpected graph generated"
        proxy = tasks[0]

        click.echo("  Cache")
        click.echo("    Identity          {0}".format(proxy.identity))
        if acache.is_available_locally(proxy):
            click.echo("    Local             {0} ({1})".format(
                True, utils.as_human_size(acache.get_artifact(proxy).get_size())))
            click.echo("    Remote            {0}".format(acache.is_available_remotely(proxy)))
            click.echo()

        if influence:
            click.echo("  Influence")
            for string in HashInfluenceRegistry.get().get_strings(task):
                click.echo("    " + string)
                click.echo()
